"""
Difference Scorer
-----------------

This scorer simply computes the elementwise difference
between two series. If the two series are multivariate, it
returns a multivariate series.
"""

import numpy as np

from darts.ad.scorers.scorers import AnomalyScorer


class DifferenceScorer(AnomalyScorer):
    def __init__(self) -> None:
        """Difference Scorer"""
        super().__init__(is_univariate=False, window=1)

    def __str__(self):
        return "Difference"

    def _score_core_from_prediction(
        self,
        vals: np.ndarray,
        pred_vals: np.ndarray,
    ) -> np.ndarray:
        vals = self._extract_deterministic_values(vals, "series")
        pred_vals = self._extract_deterministic_values(pred_vals, "pred_series")
        return vals - pred_vals



from itertools import product

import numpy as np
import pandas as pd
import pytest

from darts import concatenate
from darts.metrics import rmse
from darts.tests.conftest import TORCH_AVAILABLE, tfm_kwargs
from darts.utils import timeseries_generation as tg

if not TORCH_AVAILABLE:
    pytest.skip(
        f"Torch not available. {__name__} tests will be skipped.",
        allow_module_level=True,
    )
import torch

from darts.models.forecasting.dlinear import DLinearModel
from darts.models.forecasting.nlinear import NLinearModel
from darts.utils.likelihood_models import GaussianLikelihood


class TestDlinearNlinearModels:
    np.random.seed(42)
    torch.manual_seed(42)

    def test_creation(self):
        with pytest.raises(ValueError):
            DLinearModel(
                input_chunk_length=1,
                output_chunk_length=1,
                normalize=True,
                likelihood=GaussianLikelihood(),
            )

        with pytest.raises(ValueError):
            NLinearModel(
                input_chunk_length=1,
                output_chunk_length=1,
                normalize=True,
                likelihood=GaussianLikelihood(),
            )

    def test_fit(self):
        large_ts = tg.constant_timeseries(length=100, value=1000)
        small_ts = tg.constant_timeseries(length=100, value=10)

        for model_cls, kwargs in [
            (DLinearModel, {"kernel_size": 5}),
            (DLinearModel, {"kernel_size": 6}),
            (NLinearModel, {}),
        ]:
            # Test basic fit and predict
            model = model_cls(
                input_chunk_length=1,
                output_chunk_length=1,
                n_epochs=10,
                random_state=42,
                **kwargs,
                **tfm_kwargs,
            )
            model.fit(large_ts[:98])
            pred = model.predict(n=2).values()[0]

            # Test whether model trained on one series is better than one trained on another
            model2 = model_cls(
                input_chunk_length=1,
                output_chunk_length=1,
                n_epochs=10,
                random_state=42,
                **tfm_kwargs,
            )
            model2.fit(small_ts[:98])
            pred2 = model2.predict(n=2).values()[0]
            assert abs(pred2 - 10) < abs(pred - 10)

            # test short predict
            pred3 = model2.predict(n=1)
            assert len(pred3) == 1

    def test_logtensorboard(self, tmpdir_module):
        ts = tg.constant_timeseries(length=50, value=10)

        for model_cls in [DLinearModel, NLinearModel]:
            # Test basic fit and predict
            model = model_cls(
                input_chunk_length=1,
                output_chunk_length=1,
                n_epochs=1,
                log_tensorboard=True,
                work_dir=tmpdir_module,
                pl_trainer_kwargs={
                    "log_every_n_steps": 1,
                    **tfm_kwargs["pl_trainer_kwargs"],
                },
            )
            model.fit(ts)
            model.predict(n=2)

    def test_shared_weights(self):
        ts = tg.constant_timeseries(length=50, value=10).stack(
            tg.gaussian_timeseries(length=50)
        )

        for model_cls in [DLinearModel, NLinearModel]:
            # Test basic fit and predict
            model_shared = model_cls(
                input_chunk_length=5,
                output_chunk_length=1,
                n_epochs=2,
                const_init=False,
                shared_weights=True,
                random_state=42,
                **tfm_kwargs,
            )
            model_not_shared = model_cls(
                input_chunk_length=5,
                output_chunk_length=1,
                n_epochs=2,
                const_init=False,
                shared_weights=False,
                random_state=42,
                **tfm_kwargs,
            )
            model_shared.fit(ts)
            model_not_shared.fit(ts)
            pred_shared = model_shared.predict(n=2)
            pred_not_shared = model_not_shared.predict(n=2)
            assert np.any(np.not_equal(pred_shared.values(), pred_not_shared.values()))

    def test_multivariate_and_covariates(self):
        np.random.seed(42)
        torch.manual_seed(42)
        # test on multiple multivariate series with future and static covariates

        def _create_multiv_series(f1, f2, n1, n2, nf1, nf2):
            bases = [
                tg.sine_timeseries(length=400, value_frequency=f, value_amplitude=1.0)
                for f in (f1, f2)
            ]
            noises = [tg.gaussian_timeseries(length=400, std=n) for n in (n1, n2)]
            noise_modulators = [
                tg.sine_timeseries(length=400, value_frequency=nf)
                + tg.constant_timeseries(length=400, value=1) / 2
                for nf in (nf1, nf2)
            ]
            noises = [noises[i] * noise_modulators[i] for i in range(len(noises))]

            target = concatenate(
                [bases[i] + noises[i] for i in range(len(bases))], axis="component"
            )

            target = target.with_static_covariates(
                pd.DataFrame([[f1, n1, nf1], [f2, n2, nf2]])
            )

            return target, concatenate(noise_modulators, axis="component")

        def _eval_model(
            train1,
            train2,
            val1,
            val2,
            fut_cov1,
            fut_cov2,
            past_cov1=None,
            past_cov2=None,
            val_past_cov1=None,
            val_past_cov2=None,
            cls=DLinearModel,
            lkl=None,
            **kwargs,
        ):
            model = cls(
                input_chunk_length=50,
                output_chunk_length=10,
                shared_weights=False,
                const_init=True,
                likelihood=lkl,
                random_state=42,
                **tfm_kwargs,
            )

            model.fit(
                [train1, train2],
                past_covariates=(
                    [past_cov1, past_cov2] if past_cov1 is not None else None
                ),
                val_past_covariates=(
                    [val_past_cov1, val_past_cov2]
                    if val_past_cov1 is not None
                    else None
                ),
                future_covariates=(
                    [fut_cov1, fut_cov2] if fut_cov1 is not None else None
                ),
                epochs=10,
            )

            pred1, pred2 = model.predict(
                series=[train1, train2],
                future_covariates=(
                    [fut_cov1, fut_cov2] if fut_cov1 is not None else None
                ),
                past_covariates=(
                    [fut_cov1, fut_cov2] if past_cov1 is not None else None
                ),
                n=len(val1),
                num_samples=500 if lkl is not None else 1,
            )

            return rmse(val1, pred1), rmse(val2, pred2)

        series1, fut_cov1 = _create_multiv_series(0.05, 0.07, 0.2, 0.4, 0.02, 0.03)
        series2, fut_cov2 = _create_multiv_series(0.04, 0.03, 0.4, 0.1, 0.02, 0.04)

        train1, val1 = series1.split_after(0.7)
        train2, val2 = series2.split_after(0.7)
        past_cov1 = train1.copy()
        past_cov2 = train2.copy()
        val_past_cov1 = val1.copy()
        val_past_cov2 = val2.copy()

        for model, lkl in product(
            [DLinearModel, NLinearModel], [None, GaussianLikelihood()]
        ):
            e1, e2 = _eval_model(
                train1, train2, val1, val2, fut_cov1, fut_cov2, cls=model, lkl=lkl
            )
            assert e1 <= 0.34
            assert e2 <= 0.28

            e1, e2 = _eval_model(
                train1.with_static_covariates(None),
                train2.with_static_covariates(None),
                val1,
                val2,
                fut_cov1,
                fut_cov2,
                cls=model,
                lkl=lkl,
            )
            assert e1 <= 0.32
            assert e2 <= 0.28

            e1, e2 = _eval_model(
                train1, train2, val1, val2, None, None, cls=model, lkl=lkl
            )
            assert e1 <= 0.40
            assert e2 <= 0.34

            e1, e2 = _eval_model(
                train1.with_static_covariates(None),
                train2.with_static_covariates(None),
                val1,
                val2,
                None,
                None,
                cls=model,
                lkl=lkl,
            )
            assert e1 <= 0.40
            assert e2 <= 0.34

        e1, e2 = _eval_model(
            train1,
            train2,
            val1,
            val2,
            fut_cov1,
            fut_cov2,
            past_cov1=past_cov1,
            past_cov2=past_cov2,
            val_past_cov1=val_past_cov1,
            val_past_cov2=val_past_cov2,
            cls=NLinearModel,
            lkl=None,
            normalize=True,
        )
        # can only fit models with past/future covariates when shared_weights=False
        for model in [DLinearModel, NLinearModel]:
            for shared_weights in [True, False]:
                model_instance = model(
                    5, 5, shared_weights=shared_weights, **tfm_kwargs
                )
                assert model_instance.supports_past_covariates == (not shared_weights)
                assert model_instance.supports_future_covariates == (not shared_weights)
                if shared_weights:
                    with pytest.raises(ValueError):
                        model_instance.fit(series1, future_covariates=fut_cov1)

    def test_optional_static_covariates(self):
        series = tg.sine_timeseries(length=20).with_static_covariates(
            pd.DataFrame({"a": [1]})
        )
        for model_cls in [NLinearModel, DLinearModel]:
            # training model with static covs and predicting without will raise an error
            model = model_cls(
                input_chunk_length=12,
                output_chunk_length=6,
                use_static_covariates=True,
                n_epochs=1,
                **tfm_kwargs,
            )
            model.fit(series)
            with pytest.raises(ValueError):
                model.predict(n=2, series=series.with_static_covariates(None))

            # with `use_static_covariates=False`, static covariates are ignored and prediction works
            model = model_cls(
                input_chunk_length=12,
                output_chunk_length=6,
                use_static_covariates=False,
                n_epochs=1,
                **tfm_kwargs,
            )
            model.fit(series)
            preds = model.predict(n=2, series=series.with_static_covariates(None))
            assert preds.static_covariates is None

            # with `use_static_covariates=False`, static covariates are ignored and prediction works
            model = model_cls(
                input_chunk_length=12,
                output_chunk_length=6,
                use_static_covariates=False,
                n_epochs=1,
                **tfm_kwargs,
            )
            model.fit(series.with_static_covariates(None))
            preds = model.predict(n=2, series=series)
            assert preds.static_covariates.equals(series.static_covariates)



"""
Inference Dataset
-----------------
"""

import bisect
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
import pandas as pd
from torch.utils.data import Dataset

from darts import TimeSeries
from darts.logging import get_logger, raise_log
from darts.utils.data.utils import CovariateType
from darts.utils.historical_forecasts.utils import _process_predict_start_points_bounds

logger = get_logger(__name__)


class InferenceDataset(ABC, Dataset):
    def __init__(self):
        """
        Abstract class for all darts torch inference dataset.

        It can be used as models' inputs, to obtain simple forecasts on each `TimeSeries`
        (using covariates if specified).

        The first elements of the tuples it contains are numpy arrays (which will be translated to torch tensors
        by the torch DataLoader). The last elements of the tuples are the (past) target TimeSeries, which is
        needed in order to properly construct the time axis of the forecast series.
        """

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __getitem__(self, idx: int):
        pass

    @staticmethod
    def _covariate_indexer(
        target_idx: int,
        past_start: Union[pd.Timestamp, int],
        past_end: Union[pd.Timestamp, int],
        covariate_series: TimeSeries,
        covariate_type: CovariateType,
        input_chunk_length: int,
        output_chunk_length: int,
        output_chunk_shift: int,
        n: int,
    ):
        """returns tuple of (past_start, past_end, future_start, future_end)"""
        # get the main covariate type: CovariateType.PAST or CovariateType.FUTURE
        main_covariate_type = (
            CovariateType.PAST
            if covariate_type is CovariateType.PAST
            else CovariateType.FUTURE
        )

        if main_covariate_type not in [CovariateType.PAST, CovariateType.FUTURE]:
            raise_log(
                ValueError(
                    "`main_covariate_type` must be one of `(CovariateType.PAST, CovariateType.FUTURE)`"
                ),
                logger=logger,
            )

        # we need to use the time index (datetime or integer) here to match the index with the covariate series
        if main_covariate_type is CovariateType.PAST:
            future_end = (
                past_end + max(0, n - output_chunk_length) * covariate_series.freq
            )
        else:  # CovariateType.FUTURE
            # optionally, for future part of future covariates shift start and end by `output_chunk_shift`
            future_end = (
                past_end
                + (max(n, output_chunk_length) + output_chunk_shift)
                * covariate_series.freq
            )

        future_start = (
            past_end + covariate_series.freq * (1 + output_chunk_shift)
            if future_end != past_end
            else future_end
        )

        if input_chunk_length == 0:  # for regression ensemble models
            past_start, past_end = future_start, future_start

        # check if case specific indexes are available
        case_start = (
            future_start if covariate_type is CovariateType.FUTURE else past_start
        )
        if not covariate_series.start_time() <= case_start:
            raise_log(
                ValueError(
                    f"For the given forecasting case, the provided {main_covariate_type.value} covariates at "
                    f"dataset index `{target_idx}` do not extend far enough into the past. The "
                    f"{main_covariate_type.value} covariates must start at time step `{case_start}`, whereas now "
                    f"they start at time step `{covariate_series.start_time()}`."
                ),
                logger=logger,
            )
        if not covariate_series.end_time() >= future_end:
            raise_log(
                ValueError(
                    f"For the given forecasting horizon `n={n}`, the provided {main_covariate_type.value} covariates "
                    f"at dataset index `{target_idx}` do not extend far enough into the future. As `"
                    f"{'n > output_chunk_length' if n > output_chunk_length else 'n <= output_chunk_length'}"
                    f"` the {main_covariate_type.value} covariates must end at time step `{future_end}`, "
                    f"whereas now they end at time step `{covariate_series.end_time()}`."
                ),
                logger=logger,
            )

        # extract the index position (integer index) from time_index value
        covariate_start = covariate_series.time_index.get_loc(past_start)
        covariate_end = covariate_series.time_index.get_loc(future_end) + 1
        return covariate_start, covariate_end


class GenericInferenceDataset(InferenceDataset):
    def __init__(
        self,
        target_series: Union[TimeSeries, Sequence[TimeSeries]],
        covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        n: int = 1,
        stride: int = 0,
        bounds: Optional[np.ndarray] = None,
        input_chunk_length: int = 12,
        output_chunk_length: int = 1,
        output_chunk_shift: int = 0,
        covariate_type: CovariateType = CovariateType.PAST,
        use_static_covariates: bool = True,
    ):
        """
        Contains (past_target, past_covariates | historic_future_covariates, future_past_covariates | future_covariate,
        static_covariates).

        "future_past_covariates" are past covariates that happen to be also known in the future - those
        are needed for forecasting with n > output_chunk_length by any model relying on past covariates.
        For this reason, when n > output_chunk_length, this dataset will also emit the "future past_covariates".

        "historic_future_covariates" are historic future covariates that are given for the input_chunk in the past.

        Parameters
        ----------
        target_series
            The target series that are to be predicted into the future.
        covariates
            Optionally, one or a sequence of `TimeSeries` containing either past or future covariates. If covariates
            were used during training, the same type of cavariates must be supplied at prediction.
        n
            Forecast horizon: The number of time steps to predict after the end of the target series.
        stride
            Optionally, the number of time steps between two consecutive predictions. Can only be used together
            with `bounds`.
        bounds
            Optionally, an array of shape `(n series, 2)`, with the left and right prediction start point boundaries
            per series. The boundaries must represent the positional index of the series (0, len(series)).
            If provided, `stride` must be `>=1`.
        input_chunk_length
            The length of the target series the model takes as input.
        output_chunk_length
            The length of the target series the model emits in output.
        output_chunk_shift
            Optionally, the number of steps to shift the start of the output chunk into the future.
        use_static_covariates
            Whether to use/include static covariate data from input series.
        """
        super().__init__()

        self.target_series = (
            [target_series] if isinstance(target_series, TimeSeries) else target_series
        )
        self.covariates = (
            [covariates] if isinstance(covariates, TimeSeries) else covariates
        )

        if not (covariates is None or len(self.target_series) == len(self.covariates)):
            raise_log(
                ValueError(
                    "The number of target series must be equal to the number of covariates."
                ),
                logger=logger,
            )

        if (bounds is not None and stride == 0) or (bounds is None and stride > 0):
            raise_log(
                ValueError(
                    "Must supply either both `stride` and `bounds`, or none of them."
                ),
                logger=logger,
            )

        if output_chunk_shift and n > output_chunk_length:
            raise_log(
                ValueError(
                    "Cannot perform auto-regression `(n > output_chunk_length)` with a model that uses a "
                    "shifted output chunk `(output_chunk_shift > 0)`."
                ),
                logger=logger,
            )

        self.covariate_type = covariate_type

        self.n = n
        self.input_chunk_length = input_chunk_length
        self.output_chunk_length = output_chunk_length
        self.output_chunk_shift = output_chunk_shift
        self.use_static_covariates = use_static_covariates

        self.stride = stride
        if bounds is None:
            self.bounds = bounds
            self.cum_lengths = None
            self.len_preds = len(self.target_series)
        else:
            self.bounds, self.cum_lengths = _process_predict_start_points_bounds(
                series=target_series,
                bounds=bounds,
                stride=stride,
            )
            self.len_preds = self.cum_lengths[-1]

    def __len__(self):
        return self.len_preds

    @staticmethod
    def find_list_index(index, cumulative_lengths, bounds, stride):
        list_index = bisect.bisect_right(cumulative_lengths, index)
        bound_left = bounds[list_index, 0]
        if list_index == 0:
            stride_idx = index * stride
        else:
            stride_idx = (index - cumulative_lengths[list_index - 1]) * stride
        return list_index, bound_left + stride_idx

    def __getitem__(
        self, idx: int
    ) -> tuple[
        np.ndarray,
        Optional[np.ndarray],
        Optional[np.ndarray],
        Optional[np.ndarray],
        TimeSeries,
        Union[pd.Timestamp, int],
    ]:
        if self.bounds is None:
            series_idx, target_start_idx, target_end_idx = (
                idx,
                -self.input_chunk_length,
                None,
            )
        else:
            series_idx, target_end_idx = self.find_list_index(
                idx,
                self.cum_lengths,
                self.bounds,
                self.stride,
            )
            target_start_idx = target_end_idx - self.input_chunk_length

        target_series = self.target_series[series_idx]
        if not len(target_series) >= self.input_chunk_length:
            raise_log(
                ValueError(
                    f"All input series must have length >= `input_chunk_length` ({self.input_chunk_length})."
                ),
                logger=logger,
            )

        # extract past target values
        past_end = target_series.time_index[
            target_end_idx - 1 if target_end_idx is not None else -1
        ]
        past_target = target_series.random_component_values(copy=False)[
            target_start_idx:target_end_idx
        ]

        # optionally, extract covariates
        past_covariate, future_covariate = None, None
        covariate_series = (
            None if self.covariates is None else self.covariates[series_idx]
        )
        if covariate_series is not None:
            # get start and end indices (integer) of the covariates including historic and future parts
            covariate_start, covariate_end = self._covariate_indexer(
                target_idx=series_idx,
                past_start=target_series.time_index[target_start_idx],
                past_end=past_end,
                covariate_series=covariate_series,
                covariate_type=self.covariate_type,
                input_chunk_length=self.input_chunk_length,
                output_chunk_length=self.output_chunk_length,
                output_chunk_shift=self.output_chunk_shift,
                n=self.n,
            )

            # extract covariate values and split into a past (historic) and future part
            covariate = covariate_series.random_component_values(copy=False)[
                covariate_start:covariate_end
            ]
            if self.input_chunk_length != 0:  # regular models
                past_covariate, future_covariate = (
                    covariate[: self.input_chunk_length],
                    covariate[self.input_chunk_length + self.output_chunk_shift :],
                )
            else:  # regression ensemble models have a input_chunk_length == 0 part for using predictions as input
                past_covariate, future_covariate = covariate, covariate

            # set to None if empty array
            past_covariate = (
                past_covariate
                if past_covariate is not None and len(past_covariate) > 0
                else None
            )
            future_covariate = (
                future_covariate
                if future_covariate is not None and len(future_covariate) > 0
                else None
            )

        if self.use_static_covariates:
            static_covariate = target_series.static_covariates_values(copy=False)
        else:
            static_covariate = None

        return (
            past_target,
            past_covariate,
            future_covariate,
            static_covariate,
            target_series,
            past_end + target_series.freq * (1 + self.output_chunk_shift),
        )


class PastCovariatesInferenceDataset(InferenceDataset):
    def __init__(
        self,
        target_series: Union[TimeSeries, Sequence[TimeSeries]],
        covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        n: int = 1,
        stride: int = 0,
        bounds: Optional[np.ndarray] = None,
        input_chunk_length: int = 12,
        output_chunk_length: int = 1,
        output_chunk_shift: int = 0,
        covariate_type: CovariateType = CovariateType.PAST,
        use_static_covariates: bool = True,
    ):
        """
        Contains (past_target, past_covariates, future_past_covariates, static_covariates).

        "future_past_covariates" are past covariates that happen to be also known in the future - those
        are needed for forecasting with n > output_chunk_length by any model relying on past covariates.

        For this reason, when n > output_chunk_length, this dataset will also emit the "future past_covariates".

        Parameters
        ----------
        target_series
            The target series that are to be predicted into the future.
        covariates
            Optionally, some past-observed covariates that are used for predictions. This argument is required
            if the model was trained with past-observed covariates.
        n
            Forecast horizon: The number of time steps to predict after the end of the target series.
        stride
            Optionally, the number of time steps between two consecutive predictions. Can only be used together
            with `bounds`.
        bounds
            Optionally, an array of shape `(n series, 2)`, with the left and right prediction start point boundaries
            per series. The boundaries must represent the positional index of the series (0, len(series)).
            If provided, `stride` must be `>=1`.
        input_chunk_length
            The length of the target series the model takes as input.
        output_chunk_length
            The length of the target series the model emits in output.
        output_chunk_shift
            Optionally, the number of steps to shift the start of the output chunk into the future.
        use_static_covariates
            Whether to use/include static covariate data from input series.
        """

        super().__init__()

        self.ds = GenericInferenceDataset(
            target_series=target_series,
            covariates=covariates,
            n=n,
            stride=stride,
            bounds=bounds,
            input_chunk_length=input_chunk_length,
            output_chunk_length=output_chunk_length,
            output_chunk_shift=output_chunk_shift,
            covariate_type=covariate_type,
            use_static_covariates=use_static_covariates,
        )

    def __len__(self):
        return len(self.ds)

    def __getitem__(
        self, idx: int
    ) -> tuple[
        np.ndarray,
        Optional[np.ndarray],
        Optional[np.ndarray],
        Optional[np.ndarray],
        TimeSeries,
        Union[pd.Timestamp, int],
    ]:
        return self.ds[idx]


class FutureCovariatesInferenceDataset(InferenceDataset):
    def __init__(
        self,
        target_series: Union[TimeSeries, Sequence[TimeSeries]],
        covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        n: int = 1,
        stride: int = 0,
        bounds: Optional[np.ndarray] = None,
        input_chunk_length: int = 12,
        output_chunk_length: Optional[int] = None,
        output_chunk_shift: int = 0,
        covariate_type: CovariateType = CovariateType.FUTURE,
        use_static_covariates: bool = True,
    ):
        """
        Contains (past_target, future_covariates, static_covariates) tuples

        Parameters
        ----------
        target_series
            The target series that are to be predicted into the future.
        covariates
            Optionally, some future-known covariates that are used for predictions. This argument is required
            if the model was trained with future-known covariates.
        n
            Forecast horizon: The number of time steps to predict after the end of the target series.
        stride
            Optionally, the number of time steps between two consecutive predictions. Can only be used together
            with `bounds`.
        bounds
            Optionally, an array of shape `(n series, 2)`, with the left and right prediction start point boundaries
            per series. The boundaries must represent the positional index of the series (0, len(series)).
            If provided, `stride` must be `>=1`.
        input_chunk_length
            The length of the target series the model takes as input.
        output_chunk_length
            Optionally, the length of the target series the model emits in output. If `None`, will use the same value
            as `n`.
        output_chunk_shift
            Optionally, the number of steps to shift the start of the output chunk into the future.
        use_static_covariates
            Whether to use/include static covariate data from input series.
        """
        super().__init__()

        self.ds = GenericInferenceDataset(
            target_series=target_series,
            covariates=covariates,
            n=n,
            stride=stride,
            bounds=bounds,
            input_chunk_length=input_chunk_length,
            output_chunk_length=output_chunk_length or n,
            output_chunk_shift=output_chunk_shift,
            covariate_type=covariate_type,
            use_static_covariates=use_static_covariates,
        )

    def __len__(self):
        return len(self.ds)

    def __getitem__(
        self, idx: int
    ) -> tuple[
        np.ndarray,
        Optional[np.ndarray],
        Optional[np.ndarray],
        TimeSeries,
        Union[pd.Timestamp, int],
    ]:
        (
            past_target,
            _,
            future_covariate,
            static_covariate,
            target_series,
            pred_point,
        ) = self.ds[idx]
        return (
            past_target,
            future_covariate,
            static_covariate,
            target_series,
            pred_point,
        )


class DualCovariatesInferenceDataset(InferenceDataset):
    def __init__(
        self,
        target_series: Union[TimeSeries, Sequence[TimeSeries]],
        covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        n: int = 1,
        stride: int = 0,
        bounds: Optional[np.ndarray] = None,
        input_chunk_length: int = 12,
        output_chunk_length: int = 1,
        output_chunk_shift: int = 0,
        use_static_covariates: bool = True,
    ):
        """
        Contains (past_target, historic_future_covariates, future_covariates, static_covariates) tuples.

        Parameters
        ----------
        target_series
            The target series that are to be predicted into the future.
        covariates
            Optionally, some future-known covariates that are used for predictions. This argument is required
            if the model was trained with future-known covariates.
        n
            Forecast horizon: The number of time steps to predict after the end of the target series.
        stride
            Optionally, the number of time steps between two consecutive predictions. Can only be used together
            with `bounds`.
        bounds
            Optionally, an array of shape `(n series, 2)`, with the left and right prediction start point boundaries
            per series. The boundaries must represent the positional index of the series (0, len(series)).
            If provided, `stride` must be `>=1`.
        input_chunk_length
            The length of the target series the model takes as input.
        output_chunk_length
            The length of the target series the model emits in output.
        output_chunk_shift
            Optionally, the number of steps to shift the start of the output chunk into the future.
        use_static_covariates
            Whether to use/include static covariate data from input series.
        """
        super().__init__()

        # This dataset is in charge of serving historic future covariates
        self.ds_past = PastCovariatesInferenceDataset(
            target_series=target_series,
            covariates=covariates,
            n=n,
            stride=stride,
            bounds=bounds,
            input_chunk_length=input_chunk_length,
            output_chunk_length=output_chunk_length,
            output_chunk_shift=output_chunk_shift,
            covariate_type=CovariateType.HISTORIC_FUTURE,
            use_static_covariates=use_static_covariates,
        )

        # This dataset is in charge of serving future covariates
        self.ds_future = FutureCovariatesInferenceDataset(
            target_series=target_series,
            covariates=covariates,
            n=n,
            stride=stride,
            bounds=bounds,
            input_chunk_length=input_chunk_length,
            output_chunk_length=output_chunk_length,
            output_chunk_shift=output_chunk_shift,
            covariate_type=CovariateType.FUTURE,
            use_static_covariates=use_static_covariates,
        )

    def __len__(self):
        return len(self.ds_past)

    def __getitem__(
        self, idx
    ) -> tuple[
        np.ndarray,
        Optional[np.ndarray],
        Optional[np.ndarray],
        Optional[np.ndarray],
        TimeSeries,
        Union[pd.Timestamp, int],
    ]:
        (
            past_target,
            historic_future_covariate,
            _,
            static_covariate,
            ts_target,
            pred_point,
        ) = self.ds_past[idx]
        _, future_covariate, _, _, _ = self.ds_future[idx]
        return (
            past_target,
            historic_future_covariate,
            future_covariate,
            static_covariate,
            ts_target,
            pred_point,
        )


class MixedCovariatesInferenceDataset(InferenceDataset):
    def __init__(
        self,
        target_series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        n: int = 1,
        stride: int = 0,
        bounds: Optional[np.ndarray] = None,
        input_chunk_length: int = 12,
        output_chunk_length: int = 1,
        output_chunk_shift: int = 0,
        use_static_covariates: bool = True,
    ):
        """
        Contains (past_target, past_covariates, historic_future_covariates, future_covariates, future_past_covariates,
        static_covariates)
        tuples. "future_past_covariates" are past covariates that happen to be also known in the future - those
        are needed for forecasting with n > output_chunk_length by any model relying on past covariates.

        Parameters
        ----------
        target_series
            The target series that are to be predicted into the future.
        past_covariates
            Optionally, some past-observed covariates that are used for predictions. This argument is required
            if the model was trained with past-observed covariates.
        future_covariates
            Optionally, some future-known covariates that are used for predictions. This argument is required
            if the model was trained with future-known covariates.
        n
            Forecast horizon: The number of time steps to predict after the end of the target series.
        stride
            Optionally, the number of time steps between two consecutive predictions. Can only be used together
            with `bounds`.
        bounds
            Optionally, an array of shape `(n series, 2)`, with the left and right prediction start point boundaries
            per series. The boundaries must represent the positional index of the series (0, len(series)).
            If provided, `stride` must be `>=1`.
        input_chunk_length
            The length of the target series the model takes as input.
        output_chunk_length
            The length of the target series the model emits in output.
        output_chunk_shift
            Optionally, the number of steps to shift the start of the output chunk into the future.
        use_static_covariates
            Whether to use/include static covariate data from input series.
        """
        super().__init__()

        # This dataset is in charge of serving past covariates
        self.ds_past = PastCovariatesInferenceDataset(
            target_series=target_series,
            covariates=past_covariates,
            n=n,
            stride=stride,
            bounds=bounds,
            input_chunk_length=input_chunk_length,
            output_chunk_length=output_chunk_length,
            output_chunk_shift=output_chunk_shift,
            covariate_type=CovariateType.PAST,
            use_static_covariates=use_static_covariates,
        )

        # This dataset is in charge of serving historic and future covariates
        self.ds_future = DualCovariatesInferenceDataset(
            target_series=target_series,
            covariates=future_covariates,
            n=n,
            stride=stride,
            bounds=bounds,
            input_chunk_length=input_chunk_length,
            output_chunk_length=output_chunk_length,
            output_chunk_shift=output_chunk_shift,
            use_static_covariates=use_static_covariates,
        )

    def __len__(self):
        return len(self.ds_past)

    def __getitem__(
        self, idx
    ) -> tuple[
        np.ndarray,
        Optional[np.ndarray],
        Optional[np.ndarray],
        Optional[np.ndarray],
        Optional[np.ndarray],
        Optional[np.ndarray],
        TimeSeries,
        Union[pd.Timestamp, int],
    ]:
        (
            past_target,
            past_covariate,
            future_past_covariate,
            static_covariate,
            ts_target,
            pred_point,
        ) = self.ds_past[idx]
        _, historic_future_covariate, future_covariate, _, _, _ = self.ds_future[idx]
        return (
            past_target,
            past_covariate,
            historic_future_covariate,
            future_covariate,
            future_past_covariate,
            static_covariate,
            ts_target,
            pred_point,
        )


class SplitCovariatesInferenceDataset(InferenceDataset):
    def __init__(
        self,
        target_series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        n: int = 1,
        stride: int = 0,
        bounds: Optional[np.ndarray] = None,
        input_chunk_length: int = 12,
        output_chunk_length: int = 1,
        output_chunk_shift: int = 0,
        use_static_covariates: bool = True,
    ):
        """
        Contains (past_target, past_covariates, future_covariates, future_past_covariates, static_covariates) tuples.
        "future_past_covariates" are past covariates that happen to be also known in the future - those
        are needed for forecasting with n > output_chunk_length by any model relying on past covariates.

        Parameters
        ----------
        target_series
            The target series that are to be predicted into the future.
        past_covariates
            Optionally, some past-observed covariates that are used for predictions. This argument is required
            if the model was trained with past-observed covariates.
        future_covariates
            Optionally, some future-known covariates that are used for predictions. This argument is required
            if the model was trained with future-known covariates.
        n
            Forecast horizon: The number of time steps to predict after the end of the target series.
        stride
            Optionally, the number of time steps between two consecutive predictions. Can only be used together
            with `bounds`.
        bounds
            Optionally, an array of shape `(n series, 2)`, with the left and right prediction start point boundaries
            per series. The boundaries must represent the positional index of the series (0, len(series)).
            If provided, `stride` must be `>=1`.
        input_chunk_length
            The length of the target series the model takes as input.
        output_chunk_length
            The length of the target series the model emits in output.
        output_chunk_shift
            Optionally, the number of steps to shift the start of the output chunk into the future.
        use_static_covariates
            Whether to use/include static covariate data from input series.
        """
        super().__init__()

        # This dataset is in charge of serving past covariates
        self.ds_past = PastCovariatesInferenceDataset(
            target_series=target_series,
            covariates=past_covariates,
            n=n,
            stride=stride,
            bounds=bounds,
            input_chunk_length=input_chunk_length,
            output_chunk_length=output_chunk_length,
            output_chunk_shift=output_chunk_shift,
            covariate_type=CovariateType.PAST,
            use_static_covariates=use_static_covariates,
        )

        # This dataset is in charge of serving future covariates
        self.ds_future = FutureCovariatesInferenceDataset(
            target_series=target_series,
            covariates=future_covariates,
            n=n,
            stride=stride,
            bounds=bounds,
            input_chunk_length=input_chunk_length,
            output_chunk_length=output_chunk_length,
            output_chunk_shift=output_chunk_shift,
            covariate_type=CovariateType.FUTURE,
            use_static_covariates=use_static_covariates,
        )

    def __len__(self):
        return len(self.ds_past)

    def __getitem__(
        self, idx
    ) -> tuple[
        np.ndarray,
        Optional[np.ndarray],
        Optional[np.ndarray],
        Optional[np.ndarray],
        Optional[np.ndarray],
        TimeSeries,
        Union[pd.Timestamp, int],
    ]:
        (
            past_target,
            past_covariate,
            future_past_covariate,
            static_covariate,
            ts_target,
            pred_point,
        ) = self.ds_past[idx]
        _, future_covariate, _, _, _ = self.ds_future[idx]
        return (
            past_target,
            past_covariate,
            future_covariate,
            future_past_covariate,
            static_covariate,
            ts_target,
            pred_point,
        )



import copy
import itertools
import math
import os

import numpy as np
import pandas as pd
import pytest

from darts import TimeSeries, concatenate
from darts.datasets import AirPassengersDataset
from darts.metrics import ae, err, ic, incs_qr, mic
from darts.models import (
    ConformalNaiveModel,
    ConformalQRModel,
    LinearRegressionModel,
    NaiveSeasonal,
    NLinearModel,
)
from darts.models.forecasting.conformal_models import _get_calibration_hfc_start
from darts.models.forecasting.forecasting_model import ForecastingModel
from darts.tests.conftest import TORCH_AVAILABLE, tfm_kwargs
from darts.utils import n_steps_between
from darts.utils import timeseries_generation as tg
from darts.utils.timeseries_generation import linear_timeseries
from darts.utils.utils import (
    likelihood_component_names,
    quantile_interval_names,
    quantile_names,
)

IN_LEN = 3
OUT_LEN = 3
regr_kwargs = {"lags": IN_LEN, "output_chunk_length": OUT_LEN}
tfm_kwargs = copy.deepcopy(tfm_kwargs)
tfm_kwargs["pl_trainer_kwargs"]["fast_dev_run"] = True
torch_kwargs = dict(
    {"input_chunk_length": IN_LEN, "output_chunk_length": OUT_LEN, "random_state": 0},
    **tfm_kwargs,
)
pred_lklp = {"num_samples": 1, "predict_likelihood_parameters": True}
q = [0.1, 0.5, 0.9]


def train_model(
    *args, model_type="regression", model_params=None, quantiles=None, **kwargs
):
    model_params = model_params or {}
    if model_type == "regression":
        return LinearRegressionModel(
            **regr_kwargs,
            **model_params,
            random_state=42,
        ).fit(*args, **kwargs)
    elif model_type in ["regression_prob", "regression_qr"]:
        return LinearRegressionModel(
            likelihood="quantile",
            quantiles=quantiles,
            **regr_kwargs,
            **model_params,
            random_state=42,
        ).fit(*args, **kwargs)
    else:
        return NLinearModel(**torch_kwargs, **model_params).fit(*args, **kwargs)


# pre-trained global model for conformal models
models_cls_kwargs_errs = [
    (
        ConformalNaiveModel,
        {"quantiles": q},
        "regression",
    ),
]

if TORCH_AVAILABLE:
    models_cls_kwargs_errs.append((
        ConformalNaiveModel,
        {"quantiles": q},
        "torch",
    ))


class TestConformalModel:
    """
    Tests all general model behavior for Naive Conformal Model with symmetric non-conformity score.
    Additionally, checks correctness of predictions for:
    - ConformalNaiveModel with symmetric & asymmetric non-conformity scores
    - ConformalQRModel with symmetric & asymmetric non-conformity scores
    """

    np.random.seed(42)

    # forecasting horizon used in runnability tests
    horizon = OUT_LEN + 1

    # some arbitrary static covariates
    static_covariates = pd.DataFrame([[0.0, 1.0]], columns=["st1", "st2"])

    # real timeseries for functionality tests
    ts_length = 13 + horizon
    ts_passengers = (
        AirPassengersDataset()
        .load()[:ts_length]
        .with_static_covariates(static_covariates)
    )
    ts_pass_train, ts_pass_val = (
        ts_passengers[:-horizon],
        ts_passengers[-horizon:],
    )

    # an additional noisy series
    ts_pass_train_1 = ts_pass_train + 0.01 * tg.gaussian_timeseries(
        length=len(ts_pass_train),
        freq=ts_pass_train.freq_str,
        start=ts_pass_train.start_time(),
    )

    # an additional time series serving as covariates
    year_series = tg.datetime_attribute_timeseries(ts_passengers, attribute="year")
    month_series = tg.datetime_attribute_timeseries(ts_passengers, attribute="month")
    time_covariates = year_series.stack(month_series)
    time_covariates_train = time_covariates[:-horizon]

    # various ts with different static covariates representations
    ts_w_static_cov = tg.linear_timeseries(length=ts_length).with_static_covariates(
        pd.Series([1, 2])
    )
    ts_shared_static_cov = ts_w_static_cov.stack(tg.sine_timeseries(length=ts_length))
    ts_comps_static_cov = ts_shared_static_cov.with_static_covariates(
        pd.DataFrame([[0, 1], [2, 3]], columns=["st1", "st2"])
    )

    def test_model_construction_naive(self):
        local_model = NaiveSeasonal(K=5)
        global_model = LinearRegressionModel(**regr_kwargs)
        series = self.ts_pass_train

        model_err_msg = "`model` must be a pre-trained `GlobalForecastingModel`."
        # un-trained local model
        with pytest.raises(ValueError) as exc:
            ConformalNaiveModel(model=local_model, quantiles=q)
        assert str(exc.value) == model_err_msg

        # pre-trained local model
        local_model.fit(series)
        with pytest.raises(ValueError) as exc:
            ConformalNaiveModel(model=local_model, quantiles=q)
        assert str(exc.value) == model_err_msg

        # un-trained global model
        with pytest.raises(ValueError) as exc:
            ConformalNaiveModel(model=global_model, quantiles=q)
        assert str(exc.value) == model_err_msg

        # pre-trained local model should work
        global_model.fit(series)
        model = ConformalNaiveModel(model=global_model, quantiles=q)
        assert model.likelihood == "quantile"

        # non-centered quantiles
        with pytest.raises(ValueError) as exc:
            ConformalNaiveModel(model=global_model, quantiles=[0.2, 0.5, 0.6])
        assert str(exc.value) == (
            "quantiles lower than `q=0.5` need to share same difference to `0.5` as quantiles higher than `q=0.5`"
        )

        # quantiles missing median
        with pytest.raises(ValueError) as exc:
            ConformalNaiveModel(model=global_model, quantiles=[0.1, 0.9])
        assert str(exc.value) == "median quantile `q=0.5` must be in `quantiles`"

        # too low and high quantiles
        with pytest.raises(ValueError) as exc:
            ConformalNaiveModel(model=global_model, quantiles=[-0.1, 0.5, 1.1])
        assert str(exc.value) == "All provided quantiles must be between 0 and 1."

        # `cal_length` must be `>=1` or `None`
        with pytest.raises(ValueError) as exc:
            ConformalNaiveModel(model=global_model, quantiles=q, cal_length=0)
        assert str(exc.value) == "`cal_length` must be `>=1` or `None`."

        # `cal_stride` must be `>=1`
        with pytest.raises(ValueError) as exc:
            ConformalNaiveModel(model=global_model, quantiles=q, cal_stride=0)
        assert str(exc.value) == "`cal_stride` must be `>=1`."

        # `num_samples` must be `>=1`
        with pytest.raises(ValueError) as exc:
            ConformalNaiveModel(model=global_model, quantiles=q, cal_num_samples=0)
        assert str(exc.value) == "`cal_num_samples` must be `>=1`."

    def test_model_hfc_stride_checks(self):
        series = self.ts_pass_train
        model = LinearRegressionModel(**regr_kwargs).fit(series)
        cp_model = ConformalNaiveModel(model=model, quantiles=q, cal_stride=2)

        expected_error_start = (
            "The provided `stride` parameter must be a round-multiple of "
            "`cal_stride=2` and `>=cal_stride`."
        )
        # `stride` must be >= `cal_stride`
        with pytest.raises(ValueError) as exc:
            cp_model.historical_forecasts(series=series, stride=1)
        assert str(exc.value).startswith(expected_error_start)

        # `stride` must be a round multiple of `cal_stride`
        with pytest.raises(ValueError) as exc:
            cp_model.historical_forecasts(series=series, stride=3)
        assert str(exc.value).startswith(expected_error_start)

        # valid stride
        _ = cp_model.historical_forecasts(series=series, stride=4)

    def test_model_construction_cqr(self):
        model_det = train_model(self.ts_pass_train, model_type="regression")
        model_prob_q = train_model(
            self.ts_pass_train, model_type="regression_prob", quantiles=q
        )
        model_prob_poisson = train_model(
            self.ts_pass_train,
            model_type="regression",
            model_params={"likelihood": "poisson"},
        )

        # deterministic global model
        with pytest.raises(ValueError) as exc:
            ConformalQRModel(model=model_det, quantiles=q)
        assert str(exc.value).startswith(
            "`model` must support probabilistic forecasting."
        )
        # probabilistic model works
        _ = ConformalQRModel(model=model_prob_q, quantiles=q)
        # works also with different likelihood
        _ = ConformalQRModel(model=model_prob_poisson, quantiles=q)

    def test_unsupported_properties(self):
        """Tests only here for coverage, maybe at some point we support these properties."""
        model = ConformalNaiveModel(train_model(self.ts_pass_train), quantiles=q)
        unsupported_properties = [
            "_model_encoder_settings",
            "extreme_lags",
            "min_train_series_length",
            "min_train_samples",
        ]
        for prop in unsupported_properties:
            with pytest.raises(NotImplementedError):
                getattr(model, prop)

    @pytest.mark.parametrize("config", models_cls_kwargs_errs)
    def test_save_model_parameters(self, config):
        # model creation parameters were saved before. check if re-created model has same params as original
        model_cls, kwargs, model_type = config
        model = model_cls(
            model=train_model(
                self.ts_pass_train, model_type=model_type, quantiles=kwargs["quantiles"]
            ),
            **kwargs,
        )
        model_fresh = model.untrained_model()
        assert model._model_params.keys() == model_fresh._model_params.keys()
        for param, val in model._model_params.items():
            if isinstance(val, ForecastingModel):
                # Conformal Models require a forecasting model as input, which has no equality
                continue
            assert val == model_fresh._model_params[param]

    @pytest.mark.parametrize(
        "config", itertools.product(models_cls_kwargs_errs, [{}, pred_lklp])
    )
    def test_save_load_model(self, tmpdir_fn, config):
        # check if save and load methods work and if loaded model creates same forecasts as original model
        (model_cls, kwargs, model_type), pred_kwargs = config
        model = model_cls(
            train_model(
                self.ts_pass_train, model_type=model_type, quantiles=kwargs["quantiles"]
            ),
            **kwargs,
        )

        # check if save and load methods work and
        # if loaded conformal model creates same forecasts as original ensemble models
        expected_suffixes = [
            ".pkl",
            ".pkl.NLinearModel.pt",
            ".pkl.NLinearModel.pt.ckpt",
        ]

        # test save
        model.save()
        model.save(os.path.join(tmpdir_fn, f"{model_cls.__name__}.pkl"))

        model_prediction = model.predict(5, **pred_kwargs)

        assert os.path.exists(tmpdir_fn)
        files = os.listdir(tmpdir_fn)
        if model_type == "torch":
            # 1 from conformal model, 2 from torch, * 2 as `save()` was called twice
            assert len(files) == 6
            for f in files:
                assert f.startswith(model_cls.__name__)
            suffix_counts = {
                suffix: sum(1 for p in os.listdir(tmpdir_fn) if p.endswith(suffix))
                for suffix in expected_suffixes
            }
            assert all(count == 2 for count in suffix_counts.values())
        else:
            assert len(files) == 2
            for f in files:
                assert f.startswith(model_cls.__name__) and f.endswith(".pkl")

        # test load
        pkl_files = []
        for filename in os.listdir(tmpdir_fn):
            if filename.endswith(".pkl"):
                pkl_files.append(os.path.join(tmpdir_fn, filename))
        for p in pkl_files:
            loaded_model = model_cls.load(p)
            assert model_prediction == loaded_model.predict(5, **pred_kwargs)

    def test_fit(self):
        model = ConformalNaiveModel(train_model(self.ts_pass_train), quantiles=q)
        assert model.model._fit_called

        # check kwargs will be passed to `model.model.fit()`
        assert model.supports_sample_weight
        model.model._fit_called = False
        model.fit(self.ts_pass_train, sample_weight="linear")
        assert model.model._fit_called

    @pytest.mark.parametrize("config", models_cls_kwargs_errs)
    def test_single_ts(self, config):
        model_cls, kwargs, model_type = config
        model = model_cls(
            train_model(
                self.ts_pass_train, model_type=model_type, quantiles=kwargs["quantiles"]
            ),
            **kwargs,
        )
        pred = model.predict(n=self.horizon, **pred_lklp)
        assert pred.n_components == self.ts_pass_train.n_components * len(
            kwargs["quantiles"]
        )
        assert not np.isnan(pred.all_values()).any().any()

        pred_fc = model.model.predict(n=self.horizon)
        assert pred_fc.time_index.equals(pred.time_index)
        # the center forecasts must be equal to the forecasting model forecast
        fc_columns = likelihood_component_names(
            self.ts_pass_val.columns, quantile_names([0.5])
        )
        np.testing.assert_array_almost_equal(
            pred[fc_columns].all_values(), pred_fc.all_values()
        )
        assert pred.static_covariates is None

        # using a different `n`, gives different results, since we can generate more residuals for the horizon
        pred1 = model.predict(n=self.horizon - 1, **pred_lklp)
        assert not pred1 == pred[: len(pred1)]

        # wrong dimension
        with pytest.raises(ValueError):
            model.predict(
                n=self.horizon,
                series=self.ts_pass_train.stack(self.ts_pass_train),
                **pred_lklp,
            )

    @pytest.mark.parametrize("config", models_cls_kwargs_errs)
    def test_multi_ts(self, config):
        model_cls, kwargs, model_type = config
        model = model_cls(
            train_model(
                [self.ts_pass_train, self.ts_pass_train_1],
                model_type=model_type,
                quantiles=kwargs["quantiles"],
            ),
            **kwargs,
        )
        with pytest.raises(ValueError):
            # when model is fit from >1 series, one must provide a series in argument
            model.predict(n=1)

        pred = model.predict(n=self.horizon, series=self.ts_pass_train, **pred_lklp)
        assert pred.n_components == self.ts_pass_train.n_components * len(
            kwargs["quantiles"]
        )
        assert not np.isnan(pred.all_values()).any().any()

        # the center forecasts must be equal to the forecasting model forecast
        fc_columns = likelihood_component_names(
            self.ts_pass_val.columns, quantile_names([0.5])
        )
        pred_fc = model.model.predict(n=self.horizon, series=self.ts_pass_train)
        assert pred_fc.time_index.equals(pred.time_index)
        np.testing.assert_array_almost_equal(
            pred[fc_columns].all_values(), pred_fc.all_values()
        )

        # check prediction for several time series
        pred_list = model.predict(
            n=self.horizon,
            series=[self.ts_pass_train, self.ts_pass_train_1],
            **pred_lklp,
        )
        pred_fc_list = model.model.predict(
            n=self.horizon,
            series=[self.ts_pass_train, self.ts_pass_train_1],
        )
        assert len(pred_list) == 2, (
            f"Model {model_cls} did not return a list of prediction"
        )
        for pred, pred_fc in zip(pred_list, pred_fc_list):
            assert pred.n_components == self.ts_pass_train.n_components * len(
                kwargs["quantiles"]
            )
            assert pred_fc.time_index.equals(pred.time_index)
            assert not np.isnan(pred.all_values()).any().any()
            np.testing.assert_array_almost_equal(
                pred_fc.all_values(),
                pred[fc_columns].all_values(),
            )

        # wrong dimension
        with pytest.raises(ValueError):
            model.predict(
                n=self.horizon,
                series=[
                    self.ts_pass_train,
                    self.ts_pass_train.stack(self.ts_pass_train),
                ],
                **pred_lklp,
            )

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [(ConformalNaiveModel, {"quantiles": [0.1, 0.5, 0.9]}, "regression")],
            [
                {"lags_past_covariates": IN_LEN},
                {"lags_future_covariates": (IN_LEN, OUT_LEN)},
                {},
            ],
        ),
    )
    def test_covariates(self, config):
        (model_cls, kwargs, model_type), covs_kwargs = config
        model_fc = LinearRegressionModel(**regr_kwargs, **covs_kwargs)
        # Here we rely on the fact that all non-Dual models currently are Past models
        if model_fc.supports_future_covariates:
            cov_name = "future_covariates"
            is_past = False
        elif model_fc.supports_past_covariates:
            cov_name = "past_covariates"
            is_past = True
        else:
            cov_name = None
            is_past = None

        covariates = [self.time_covariates_train, self.time_covariates_train]
        if cov_name is not None:
            cov_kwargs = {cov_name: covariates}
            cov_kwargs_train = {cov_name: self.time_covariates_train}
            cov_kwargs_notrain = {cov_name: self.time_covariates}
        else:
            cov_kwargs = {}
            cov_kwargs_train = {}
            cov_kwargs_notrain = {}

        model_fc.fit(series=[self.ts_pass_train, self.ts_pass_train_1], **cov_kwargs)

        model = model_cls(model=model_fc, **kwargs)
        if cov_name == "future_covariates":
            assert model.supports_future_covariates
            assert not model.supports_past_covariates
            assert model.uses_future_covariates
            assert not model.uses_past_covariates
        elif cov_name == "past_covariates":
            assert not model.supports_future_covariates
            assert model.supports_past_covariates
            assert not model.uses_future_covariates
            assert model.uses_past_covariates
        else:
            assert not model.supports_future_covariates
            assert not model.supports_past_covariates
            assert not model.uses_future_covariates
            assert not model.uses_past_covariates

        with pytest.raises(ValueError):
            # when model is fit from >1 series, one must provide a series in argument
            model.predict(n=1)

        if cov_name is not None:
            with pytest.raises(ValueError):
                # when model is fit using multiple covariates, covariates are required at prediction time
                model.predict(n=1, series=self.ts_pass_train)

            with pytest.raises(ValueError):
                # when model is fit using covariates, n cannot be greater than output_chunk_length...
                # (for short covariates)
                # past covariates model can predict up until output_chunk_length
                # with train future covariates we cannot predict at all after end of series
                model.predict(
                    n=OUT_LEN + 1 if is_past else 1,
                    series=self.ts_pass_train,
                    **cov_kwargs_train,
                )
        else:
            # model does not support covariates
            with pytest.raises(ValueError):
                model.predict(
                    n=1,
                    series=self.ts_pass_train,
                    past_covariates=self.time_covariates,
                )
            with pytest.raises(ValueError):
                model.predict(
                    n=1,
                    series=self.ts_pass_train,
                    future_covariates=self.time_covariates,
                )

        # ... unless future covariates are provided
        _ = model.predict(
            n=self.horizon, series=self.ts_pass_train, **cov_kwargs_notrain
        )

        pred = model.predict(
            n=self.horizon, series=self.ts_pass_train, **cov_kwargs_notrain, **pred_lklp
        )
        pred_fc = model_fc.predict(
            n=self.horizon,
            series=self.ts_pass_train,
            **cov_kwargs_notrain,
        )
        fc_columns = likelihood_component_names(
            self.ts_pass_val.columns, quantile_names([0.5])
        )
        np.testing.assert_array_almost_equal(
            pred[fc_columns].all_values(),
            pred_fc.all_values(),
        )

        if cov_name is None:
            return

        # when model is fit using 1 training and 1 covariate series, time series args are optional
        model_fc = LinearRegressionModel(**regr_kwargs, **covs_kwargs)
        model_fc.fit(series=self.ts_pass_train, **cov_kwargs_train)
        model = model_cls(model_fc, **kwargs)

        if is_past:
            # can only predict up until ocl
            with pytest.raises(ValueError):
                _ = model.predict(n=OUT_LEN + 1, **pred_lklp)
            # wrong covariates dimension
            with pytest.raises(ValueError):
                covs = cov_kwargs_train[cov_name]
                covs = {cov_name: covs.stack(covs)}
                _ = model.predict(n=OUT_LEN, **covs, **pred_lklp)
            # with past covariates from train we can predict up until output_chunk_length
            pred1 = model.predict(n=OUT_LEN, **pred_lklp)
            pred2 = model.predict(n=OUT_LEN, series=self.ts_pass_train, **pred_lklp)
            pred3 = model.predict(n=OUT_LEN, **cov_kwargs_train, **pred_lklp)
            pred4 = model.predict(
                n=OUT_LEN, **cov_kwargs_train, series=self.ts_pass_train, **pred_lklp
            )
        else:
            # with future covariates we need additional time steps to predict
            with pytest.raises(ValueError):
                _ = model.predict(n=1, **pred_lklp)
            with pytest.raises(ValueError):
                _ = model.predict(n=1, series=self.ts_pass_train, **pred_lklp)
            with pytest.raises(ValueError):
                _ = model.predict(n=1, **cov_kwargs_train, **pred_lklp)
            with pytest.raises(ValueError):
                _ = model.predict(
                    n=1, **cov_kwargs_train, series=self.ts_pass_train, **pred_lklp
                )
            # wrong covariates dimension
            with pytest.raises(ValueError):
                covs = cov_kwargs_notrain[cov_name]
                covs = {cov_name: covs.stack(covs)}
                _ = model.predict(n=OUT_LEN, **covs, **pred_lklp)
            pred1 = model.predict(n=OUT_LEN, **cov_kwargs_notrain, **pred_lklp)
            pred2 = model.predict(
                n=OUT_LEN, series=self.ts_pass_train, **cov_kwargs_notrain, **pred_lklp
            )
            pred3 = model.predict(n=OUT_LEN, **cov_kwargs_notrain, **pred_lklp)
            pred4 = model.predict(
                n=OUT_LEN, **cov_kwargs_notrain, series=self.ts_pass_train, **pred_lklp
            )

        assert pred1 == pred2
        assert pred1 == pred3
        assert pred1 == pred4

    @pytest.mark.parametrize(
        "config,ts",
        itertools.product(
            models_cls_kwargs_errs,
            [ts_w_static_cov, ts_shared_static_cov, ts_comps_static_cov],
        ),
    )
    def test_use_static_covariates(self, config, ts):
        """
        Check that both static covariates representations are supported (component-specific and shared)
        for both uni- and multivariate series when fitting the model.
        Also check that the static covariates are present in the forecasted series
        """
        model_cls, kwargs, model_type = config
        model = model_cls(
            train_model(ts, model_type=model_type, quantiles=kwargs["quantiles"]),
            **kwargs,
        )
        assert model.considers_static_covariates
        assert model.supports_static_covariates
        assert model.uses_static_covariates
        pred = model.predict(OUT_LEN)
        assert pred.static_covariates.equals(ts.static_covariates)

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [True, False],  # univariate series
            [True, False],  # single series
            [True, False],  # use covariates
            [True, False],  # datetime index
            [1, 3, 5],  # different horizons
        ),
    )
    def test_predict(self, config):
        (is_univar, is_single, use_covs, is_datetime, horizon) = config
        series = self.ts_pass_train
        if not is_univar:
            series = series.stack(series)
        if not is_datetime:
            series = TimeSeries.from_values(series.all_values(), columns=series.columns)
        if use_covs:
            pc, fc = series, series
            fc = fc.append_values(fc.values()[: max(horizon, OUT_LEN)])
            if horizon > OUT_LEN:
                pc = pc.append_values(pc.values()[: horizon - OUT_LEN])
            model_kwargs = {
                "lags_past_covariates": IN_LEN,
                "lags_future_covariates": (IN_LEN, OUT_LEN),
            }
        else:
            pc, fc = None, None
            model_kwargs = {}
        if not is_single:
            series = [
                series,
                series.with_columns_renamed(
                    col_names=series.columns.tolist(),
                    col_names_new=(series.columns + "_s2").tolist(),
                ),
            ]
            if use_covs:
                pc = [pc] * 2
                fc = [fc] * 2

        # testing lags_past_covariates None but past_covariates during prediction
        model_instance = LinearRegressionModel(
            lags=IN_LEN, output_chunk_length=OUT_LEN, **model_kwargs
        )
        model_instance.fit(series=series, past_covariates=pc, future_covariates=fc)
        model = ConformalNaiveModel(model_instance, quantiles=q)

        preds = model.predict(
            n=horizon,
            series=series,
            past_covariates=pc,
            future_covariates=fc,
            **pred_lklp,
        )

        if is_single:
            series = [series]
            preds = [preds]

        for s_, preds_ in zip(series, preds):
            cols_expected = likelihood_component_names(s_.columns, quantile_names(q))
            assert preds_.columns.tolist() == cols_expected
            assert len(preds_) == horizon
            assert preds_.start_time() == s_.end_time() + s_.freq
            assert preds_.freq == s_.freq

    def test_output_chunk_shift(self):
        model_params = {"output_chunk_shift": 1}
        model = ConformalNaiveModel(
            train_model(self.ts_pass_train, model_params=model_params, quantiles=q),
            quantiles=q,
        )
        pred = model.predict(n=1, **pred_lklp)
        pred_fc = model.model.predict(n=1)

        assert pred_fc.time_index.equals(pred.time_index)
        # the center forecasts must be equal to the forecasting model forecast
        fc_columns = likelihood_component_names(
            self.ts_pass_train.columns, quantile_names([0.5])
        )

        np.testing.assert_array_almost_equal(
            pred[fc_columns].all_values(), pred_fc.all_values()
        )

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [1, 3, 5],  # horizon
            [True, False],  # univariate series
            [True, False],  # single series
            [q, [0.2, 0.3, 0.5, 0.7, 0.8]],
            [
                (ConformalNaiveModel, "regression"),
                (ConformalNaiveModel, "regression_prob"),
                (ConformalQRModel, "regression_qr"),
            ],  # model type
            [True, False],  # symmetric non-conformity score
            [None, 1],  # train length
        ),
    )
    def test_conformal_model_predict_accuracy(self, config):
        """Verifies that naive conformal model computes the correct intervals for:
        - different horizons (smaller, equal, larger than ocl)
        - uni/multivariate series
        - single/multi series
        - single/multi quantile intervals
        - deterministic/probabilistic forecasting model
        - naive conformal and conformalized quantile regression
        - symmetric/asymmetric non-conformity scores

        The naive approach computes it as follows:

        - pred_upper = pred + q_interval(absolute error, past)
        - pred_middle = pred
        - pred_lower = pred - q_interval(absolute error, past)

        Where q_interval(absolute error) is the `q_hi - q_hi` quantile value of all historic absolute errors
        between `pred`, and the target series.
        """
        (
            n,
            is_univar,
            is_single,
            quantiles,
            (model_cls, model_type),
            symmetric,
            cal_length,
        ) = config
        idx_med = quantiles.index(0.5)
        q_intervals = [
            (q_hi, q_lo)
            for q_hi, q_lo in zip(quantiles[:idx_med], quantiles[idx_med + 1 :][::-1])
        ]
        series = self.helper_prepare_series(is_univar, is_single)
        pred_kwargs = (
            {"num_samples": 1000}
            if model_type in ["regression_prob", "regression_qr"]
            else {}
        )

        model_fc = train_model(series, model_type=model_type, quantiles=q)
        model = model_cls(
            model=model_fc,
            quantiles=quantiles,
            symmetric=symmetric,
            cal_length=cal_length,
        )
        pred_fc_list = model.model.predict(n, series=series, **pred_kwargs)
        pred_cal_list = model.predict(n, series=series, **pred_lklp)

        if issubclass(model_cls, ConformalNaiveModel):
            metric = ae if symmetric else err
            metric_kwargs = {}
        else:
            metric = incs_qr
            metric_kwargs = {"q_interval": q_intervals, "symmetric": symmetric}
        # compute the expected intervals
        residuals_list = model.model.residuals(
            series,
            retrain=False,
            forecast_horizon=n,
            overlap_end=True,
            last_points_only=False,
            stride=1,
            values_only=True,
            metric=metric,
            metric_kwargs=metric_kwargs,
            **pred_kwargs,
        )
        if is_single:
            pred_fc_list = [pred_fc_list]
            pred_cal_list = [pred_cal_list]
            residuals_list = [residuals_list]

        for pred_fc, pred_cal, residuals in zip(
            pred_fc_list, pred_cal_list, residuals_list
        ):
            residuals = np.concatenate(residuals[:-1], axis=2)

            pred_vals = pred_fc.all_values()
            pred_vals_expected = self.helper_compute_pred_cal(
                residuals,
                pred_vals,
                n,
                quantiles,
                model_type,
                symmetric,
                cal_length=cal_length,
            )
            self.helper_compare_preds(pred_cal, pred_vals_expected, model_type)

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [1, 3, 5],  # horizon
            [True, False],  # univariate series
            [True, False],  # single series,
            [0, 1],  # output chunk shift
            [None, 1],  # train length
            [False, True],  # use covariates
            [q, [0.2, 0.3, 0.5, 0.7, 0.8]],  # quantiles
        ),
    )
    def test_naive_conformal_model_historical_forecasts(self, config):
        """Checks correctness of naive conformal model historical forecasts for:
        - different horizons (smaller, equal and larger the OCL)
        - uni and multivariate series
        - single and multiple series
        - with and without output shift
        - with and without training length
        - with and without covariates
        """
        n, is_univar, is_single, ocs, cal_length, use_covs, quantiles = config
        if ocs and n > OUT_LEN:
            # auto-regression not allowed with ocs
            return

        series = self.helper_prepare_series(is_univar, is_single)
        model_params = {"output_chunk_shift": ocs}

        # for covariates, we check that shorter & longer covariates in the calibration set give expected results
        covs_kwargs = {}
        if use_covs:
            model_params["lags_past_covariates"] = regr_kwargs["lags"]
            past_covs = series
            if n > OUT_LEN:
                append_vals = [[[1.0]] * (1 if is_univar else 2)] * (n - OUT_LEN)
                if is_single:
                    past_covs = past_covs.append_values(append_vals)
                else:
                    past_covs = [pc.append_values(append_vals) for pc in past_covs]
            covs_kwargs["past_covariates"] = past_covs

        # forecasts from forecasting model
        model_fc = train_model(series, model_params=model_params, **covs_kwargs)
        hfc_fc_list = model_fc.historical_forecasts(
            series,
            retrain=False,
            forecast_horizon=n,
            overlap_end=True,
            last_points_only=False,
            stride=1,
            **covs_kwargs,
        )
        # residuals to compute the conformal intervals
        residuals_list = model_fc.residuals(
            series,
            historical_forecasts=hfc_fc_list,
            overlap_end=True,
            last_points_only=False,
            values_only=True,
            metric=ae,  # absolute error
            **covs_kwargs,
        )

        # conformal forecasts
        model = ConformalNaiveModel(
            model=model_fc, quantiles=quantiles, cal_length=cal_length
        )
        hfc_conf_list = model.historical_forecasts(
            series=series,
            forecast_horizon=n,
            overlap_end=True,
            last_points_only=False,
            stride=1,
            **covs_kwargs,
            **pred_lklp,
        )

        if is_single:
            hfc_conf_list = [hfc_conf_list]
            residuals_list = [residuals_list]
            hfc_fc_list = [hfc_fc_list]

        # validate computed conformal intervals; conformal models start later since they need past residuals as input
        first_fc_idx = len(hfc_fc_list[0]) - len(hfc_conf_list[0])
        for hfc_fc, hfc_conf, hfc_residuals in zip(
            hfc_fc_list, hfc_conf_list, residuals_list
        ):
            for idx, (pred_fc, pred_cal) in enumerate(
                zip(hfc_fc[first_fc_idx:], hfc_conf)
            ):
                # need to ignore additional `ocs` (output shift) residuals
                residuals = np.concatenate(
                    hfc_residuals[: first_fc_idx - ocs + idx], axis=2
                )

                pred_vals = pred_fc.all_values()
                pred_vals_expected = self.helper_compute_pred_cal(
                    residuals,
                    pred_vals,
                    n,
                    quantiles,
                    cal_length=cal_length,
                    model_type="regression",
                    symmetric=True,
                )
                np.testing.assert_array_almost_equal(
                    pred_cal.all_values(), pred_vals_expected
                )

        # checking that last points only is equal to the last forecasted point
        hfc_lpo_list = model.historical_forecasts(
            series=series,
            forecast_horizon=n,
            overlap_end=True,
            last_points_only=True,
            stride=1,
            **covs_kwargs,
            **pred_lklp,
        )
        if is_single:
            hfc_lpo_list = [hfc_lpo_list]

        for hfc_lpo, hfc_conf in zip(hfc_lpo_list, hfc_conf_list):
            hfc_conf_lpo = concatenate([hfc[-1:] for hfc in hfc_conf], axis=0)
            assert hfc_lpo == hfc_conf_lpo

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [1, 3, 5],  # horizon
            [0, 1],  # output chunk shift
            [None, 1],  # cal length,
            [1, 2],  # cal stride
            [False, True],  # use start
        ),
    )
    def test_stridden_conformal_model(self, config):
        """Checks correctness of naive conformal model historical forecasts for:
        - different horizons (smaller, equal and larger the OCL)
        - uni and multivariate series
        - single and multiple series
        - with and without output shift
        - with and without training length
        - with and without covariates
        """
        is_univar, is_single = True, False
        n, ocs, cal_length, cal_stride, use_start = config
        if ocs and n > OUT_LEN:
            # auto-regression not allowed with ocs
            return

        series = self.helper_prepare_series(is_univar, is_single)
        # shift second series ahead to cover the non overlapping multi series case
        series = [series[0], series[1].shift(120)]
        model_params = {"output_chunk_shift": ocs}

        # forecasts from forecasting model
        model_fc = train_model(series, model_params=model_params)
        hfc_fc_list = model_fc.historical_forecasts(
            series,
            retrain=False,
            forecast_horizon=n,
            overlap_end=True,
            last_points_only=False,
            stride=cal_stride,
        )
        # residuals to compute the conformal intervals
        residuals_list = model_fc.residuals(
            series,
            historical_forecasts=hfc_fc_list,
            overlap_end=True,
            last_points_only=False,
            values_only=True,
            metric=ae,  # absolute error
        )

        # conformal forecasts
        model = ConformalNaiveModel(
            model=model_fc,
            quantiles=q,
            cal_length=cal_length,
            cal_stride=cal_stride,
        )
        # the expected positional index of the first conformal forecast
        # index = (skip n + ocs points (relative to cal_stride) to avoid look-ahead bias) + (number of cal examples)
        first_fc_idx = math.ceil((n + ocs) / cal_stride) + (
            cal_length - 1 if cal_length else 0
        )
        first_start = n_steps_between(
            hfc_fc_list[0][first_fc_idx].start_time() - ocs * series[0].freq,
            series[0].start_time(),
            freq=series[0].freq,
        )

        hfc_conf_list = model.historical_forecasts(
            series=series,
            forecast_horizon=n,
            overlap_end=True,
            last_points_only=False,
            start=first_start if use_start else None,
            start_format="position" if use_start else "value",
            stride=cal_stride,
            **pred_lklp,
        )

        # also, skip some residuals from output chunk shift
        ignore_ocs = math.ceil(ocs / cal_stride) if ocs >= cal_stride else 0
        for hfc_fc, hfc_conf, hfc_residuals in zip(
            hfc_fc_list, hfc_conf_list, residuals_list
        ):
            for idx, (pred_fc, pred_cal) in enumerate(
                zip(hfc_fc[first_fc_idx:], hfc_conf)
            ):
                residuals = np.concatenate(
                    hfc_residuals[: first_fc_idx - ignore_ocs + idx], axis=2
                )
                pred_vals = pred_fc.all_values()
                pred_vals_expected = self.helper_compute_pred_cal(
                    residuals,
                    pred_vals,
                    n,
                    q,
                    cal_length=cal_length,
                    model_type="regression",
                    symmetric=True,
                    cal_stride=cal_stride,
                )
                assert pred_fc.time_index.equals(pred_cal.time_index)
                np.testing.assert_array_almost_equal(
                    pred_cal.all_values(), pred_vals_expected
                )

        # check that with a round-multiple of `cal_stride` we get identical forecasts
        assert model.historical_forecasts(
            series=series,
            forecast_horizon=n,
            overlap_end=True,
            last_points_only=False,
            start=first_start if use_start else None,
            start_format="position" if use_start else "value",
            stride=2 * cal_stride,
            **pred_lklp,
        ) == [hfc[::2] for hfc in hfc_conf_list]

        # checking that last points only is equal to the last forecasted point
        hfc_lpo_list = model.historical_forecasts(
            series=series,
            forecast_horizon=n,
            overlap_end=True,
            last_points_only=True,
            stride=cal_stride,
            **pred_lklp,
        )
        for hfc_lpo, hfc_conf in zip(hfc_lpo_list, hfc_conf_list):
            hfc_conf_lpo = concatenate(
                [hfc[-1::cal_stride] for hfc in hfc_conf], axis=0
            )
            assert hfc_lpo == hfc_conf_lpo

        # checking that predict gives the same results as last historical forecast
        preds = model.predict(
            series=series,
            n=n,
            **pred_lklp,
        )
        hfcs_conf_end = model.historical_forecasts(
            series=series,
            forecast_horizon=n,
            overlap_end=True,
            last_points_only=False,
            start=-cal_stride,
            start_format="position",
            stride=cal_stride,
            **pred_lklp,
        )
        hfcs_conf_end = [hfc[-1] for hfc in hfcs_conf_end]
        for pred, last_hfc in zip(preds, hfcs_conf_end):
            assert pred == last_hfc

    def test_probabilistic_historical_forecast(self):
        """Checks correctness of naive conformal historical forecast from probabilistic fc model compared to
        deterministic one,
        """
        series = self.helper_prepare_series(False, False)
        # forecasts from forecasting model
        model_det = ConformalNaiveModel(
            train_model(series, model_type="regression", quantiles=q),
            quantiles=q,
        )
        model_prob = ConformalNaiveModel(
            train_model(series, model_type="regression_prob", quantiles=q),
            quantiles=q,
        )
        hfcs_det = model_det.historical_forecasts(
            series,
            forecast_horizon=2,
            last_points_only=True,
            stride=1,
            **pred_lklp,
        )
        hfcs_prob = model_prob.historical_forecasts(
            series,
            forecast_horizon=2,
            last_points_only=True,
            stride=1,
            **pred_lklp,
        )
        assert isinstance(hfcs_det, list) and len(hfcs_det) == 2
        assert isinstance(hfcs_prob, list) and len(hfcs_prob) == 2
        for hfc_det, hfc_prob in zip(hfcs_det, hfcs_prob):
            assert hfc_det.columns.equals(hfc_prob.columns)
            assert hfc_det.time_index.equals(hfc_prob.time_index)
            self.helper_compare_preds(
                hfc_prob, hfc_det.all_values(), model_type="regression_prob"
            )

    def helper_prepare_series(self, is_univar, is_single):
        series = self.ts_pass_train
        if not is_univar:
            series = series.stack(series + 3.0)
        if not is_single:
            series = [series, series + 5]
        return series

    @staticmethod
    def helper_compare_preds(cp_pred, pred_expected, model_type, tol_rel=0.1):
        if isinstance(cp_pred, TimeSeries):
            cp_pred = cp_pred.all_values(copy=False)
        if model_type == "regression":
            # deterministic fc model should give almost identical results
            np.testing.assert_array_almost_equal(cp_pred, pred_expected)
        else:
            # probabilistic fc models have some randomness
            diffs_rel = np.abs((cp_pred - pred_expected) / pred_expected)
            assert (diffs_rel < tol_rel).all().all()

    @staticmethod
    def helper_compute_pred_cal(
        residuals,
        pred_vals,
        horizon,
        quantiles,
        model_type,
        symmetric,
        cal_length=None,
        cal_stride=1,
    ):
        """Generates expected prediction results for naive conformal model from:

        - residuals and predictions from deterministic/probabilistic model
        - any forecast horizon
        - any quantile intervals
        - symmetric/ asymmetric non-conformity scores
        - any train length
        """
        cal_length = cal_length or 0
        n_comps = pred_vals.shape[1]
        half_idx = len(quantiles) // 2

        # get alphas from quantiles (alpha = q_hi - q_lo) per interval
        alphas = np.array(quantiles[half_idx + 1 :][::-1]) - np.array(
            quantiles[:half_idx]
        )
        if not symmetric:
            # asymmetric non-conformity scores look only on one tail -> alpha/2
            alphas = 1 - (1 - alphas) / 2
        if model_type == "regression_prob":
            # naive conformal model converts probabilistic forecasts to median (deterministic)
            pred_vals = np.expand_dims(np.quantile(pred_vals, 0.5, axis=2), -1)
        elif model_type == "regression_qr":
            # conformalized quantile regression consumes quantile forecasts
            pred_vals = np.quantile(pred_vals, quantiles, axis=2).transpose(1, 2, 0)

        is_naive = model_type in ["regression", "regression_prob"]
        pred_expected = []
        for alpha_idx, alpha in enumerate(alphas):
            q_hats = []
            # compute the quantile `alpha` of all past residuals (absolute "per time step" errors between historical
            # forecasts and the target series)
            for idx_horizon in range(horizon):
                n = idx_horizon + 1
                # ignore residuals at beginning
                idx_fc_start = math.floor((horizon - n) / cal_stride)
                # keep as many residuals as possible from end
                idx_fc_end = -(math.ceil(horizon / cal_stride) - (idx_fc_start + 1))
                res_n = residuals[idx_horizon, :, idx_fc_start : idx_fc_end or None]
                if cal_length is not None:
                    res_n = res_n[:, -cal_length:]
                if is_naive and symmetric:
                    # identical correction for upper and lower bounds
                    # metric is `ae()`
                    q_hat_n = np.quantile(res_n, q=alpha, method="higher", axis=1)
                    q_hats.append((-q_hat_n, q_hat_n))
                elif is_naive:
                    # correction separately for upper and lower bounds
                    # metric is `err()`
                    q_hat_hi = np.quantile(res_n, q=alpha, method="higher", axis=1)
                    q_hat_lo = np.quantile(-res_n, q=alpha, method="higher", axis=1)
                    q_hats.append((-q_hat_lo, q_hat_hi))
                elif symmetric:  # CQR symmetric
                    # identical correction for upper and lower bounds
                    # metric is `incs_qr(symmetric=True)`
                    q_hat_n = np.quantile(res_n, q=alpha, method="higher", axis=1)
                    q_hats.append((-q_hat_n, q_hat_n))
                else:  # CQR asymmetric
                    # correction separately for upper and lower bounds
                    # metric is `incs_qr(symmetric=False)`
                    half_idx = len(res_n) // 2

                    # residuals have shape (n components * n intervals * 2)
                    # the factor 2 comes from the metric being computed for lower, and upper bounds separately
                    # (comp_1_qlow_1, comp_1_qlow_2, ... comp_n_qlow_m, comp_1_qhigh_1, ...)
                    q_hat_lo = np.quantile(
                        res_n[:half_idx], q=alpha, method="higher", axis=1
                    )
                    q_hat_hi = np.quantile(
                        res_n[half_idx:], q=alpha, method="higher", axis=1
                    )
                    q_hats.append((
                        -q_hat_lo[alpha_idx :: len(alphas)],
                        q_hat_hi[alpha_idx :: len(alphas)],
                    ))
            # bring to shape (horizon, n components, 2)
            q_hats = np.array(q_hats).transpose((0, 2, 1))
            # the prediction interval is given by pred +/- q_hat
            pred_vals_expected = []
            for col_idx in range(n_comps):
                q_col = q_hats[:, col_idx]
                pred_col = pred_vals[:, col_idx]
                if is_naive:
                    # conformal model corrects deterministic predictions
                    idx_q_lo = slice(0, None)
                    idx_q_med = slice(0, None)
                    idx_q_hi = slice(0, None)
                else:
                    # conformal model corrects quantile predictions
                    idx_q_lo = slice(alpha_idx, alpha_idx + 1)
                    idx_q_med = slice(len(alphas), len(alphas) + 1)
                    idx_q_hi = slice(
                        pred_col.shape[1] - (alpha_idx + 1),
                        pred_col.shape[1] - alpha_idx,
                    )
                # correct lower and upper bounds
                pred_col_expected = np.concatenate(
                    [
                        pred_col[:, idx_q_lo] + q_col[:, :1],  # lower quantile
                        pred_col[:, idx_q_med],  # median forecast
                        pred_col[:, idx_q_hi] + q_col[:, 1:],
                    ],  # upper quantile
                    axis=1,
                )
                pred_col_expected = np.expand_dims(pred_col_expected, 1)
                pred_vals_expected.append(pred_col_expected)
            pred_vals_expected = np.concatenate(pred_vals_expected, axis=1)
            pred_expected.append(pred_vals_expected)

        # reorder to have columns going from lowest quantiles to highest per component
        pred_expected_reshaped = []
        for comp_idx in range(n_comps):
            for q_idx in [0, 1, 2]:
                for pred_idx in range(len(pred_expected)):
                    # upper quantiles will have reversed order
                    if q_idx == 2:
                        pred_idx = len(pred_expected) - 1 - pred_idx
                    pred_ = pred_expected[pred_idx][:, comp_idx, q_idx]
                    pred_ = pred_.reshape(-1, 1, 1)

                    # q_hat_idx = q_idx + comp_idx * 3 + alpha_idx * 3 * n_comps
                    pred_expected_reshaped.append(pred_)
                    # only add median quantile once
                    if q_idx == 1:
                        break
        return np.concatenate(pred_expected_reshaped, axis=1)

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [1, 3, 5],  # horizon
            [0, 1],  # output chunk shift
            [False, True],  # use covariates
        ),
    )
    def test_too_short_input_predict(self, config):
        """Checks conformal model predict with minimum required input and too short input."""
        n, ocs, use_covs = config
        if ocs and n > OUT_LEN:
            return
        icl = IN_LEN
        min_len = icl + ocs + n
        series = tg.linear_timeseries(length=min_len)
        series_train = [tg.linear_timeseries(length=IN_LEN + OUT_LEN + ocs)] * 2

        model_params = {"output_chunk_shift": ocs}
        covs_kwargs = {}
        covs_kwargs_train = {}
        covs_kwargs_too_short = {}
        if use_covs:
            model_params["lags_past_covariates"] = regr_kwargs["lags"]
            covs_kwargs_train["past_covariates"] = series_train
            # use shorter covariates, to test whether residuals are still properly extracted
            past_covs = series
            # for auto-regression, we require longer past covariates
            if n > OUT_LEN:
                past_covs = past_covs.append_values([1.0] * (n - OUT_LEN))
            covs_kwargs["past_covariates"] = past_covs
            covs_kwargs_too_short["past_covariates"] = past_covs[:-1]

        model = ConformalNaiveModel(
            train_model(
                series=series_train,
                model_params=model_params,
                **covs_kwargs_train,
            ),
            quantiles=q,
        )

        # prediction works with long enough input
        preds1 = model.predict(n=n, series=series, **covs_kwargs)
        assert not np.isnan(preds1.all_values()).any().any()

        # series too short: without covariates, make `series` shorter. Otherwise, use the shorter covariates
        series_ = series[:-1] if not use_covs else series
        with pytest.raises(ValueError) as exc:
            _ = model.predict(n=n, series=series_, **covs_kwargs_too_short)
        if not use_covs:
            assert str(exc.value).startswith(
                "Could not build the minimum required calibration input with the provided `series`"
            )
        else:
            # if `past_covariates` are too short, then it raises error from the forecasting_model.predict()
            assert str(exc.value).startswith(
                "The `past_covariates` are not long enough."
            )

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [False, True],  # last points only
            [False, True],  # overlap end
            [None, 2],  # train length
            [0, 1],  # output chunk shift
            [1, 3, 5],  # horizon
            [True, False],  # use covs
        ),
    )
    def test_too_short_input_hfc(self, config):
        """Checks conformal model historical forecasts with minimum required input and too short input."""
        (
            last_points_only,
            overlap_end,
            cal_length,
            ocs,
            n,
            use_covs,
        ) = config
        if ocs and n > OUT_LEN:
            return

        icl = IN_LEN
        ocl = OUT_LEN
        horizon_ocs = n + ocs
        add_cal_length = cal_length - 1 if cal_length is not None else 0
        # min length to generate 1 conformal forecast
        min_len_val_series = (
            icl + horizon_ocs * (1 + int(not overlap_end)) + add_cal_length
        )

        series_train = [tg.linear_timeseries(length=icl + ocl + ocs)] * 2
        series = tg.linear_timeseries(length=min_len_val_series)

        model_params = {"output_chunk_shift": ocs}
        covs_kwargs_train = {}
        covs_kwargs = {}
        covs_kwargs_short = {}
        if use_covs:
            model_params["lags_past_covariates"] = regr_kwargs["lags"]
            covs_kwargs_train["past_covariates"] = series_train

            # `- horizon_ocs` to generate forecasts extending up until end of target series
            if not overlap_end:
                past_covs = series[:-horizon_ocs]
            else:
                past_covs = series

            # for auto-regression, we require longer past covariates
            if n > OUT_LEN:
                past_covs = past_covs.append_values([1.0] * (n - OUT_LEN))

            # covariates lengths to generate exactly one forecast
            covs_kwargs["past_covariates"] = past_covs

            # use too short covariates to check that errors are raised
            covs_kwargs_short["past_covariates"] = covs_kwargs["past_covariates"][:-1]

        model = ConformalNaiveModel(
            train_model(
                series=series_train,
                model_params=model_params,
                **covs_kwargs_train,
            ),
            quantiles=q,
            cal_length=cal_length,
        )

        hfc_kwargs = {
            "last_points_only": last_points_only,
            "overlap_end": overlap_end,
            "forecast_horizon": n,
        }
        # prediction works with long enough input
        hfcs = model.historical_forecasts(
            series=series,
            **covs_kwargs,
            **hfc_kwargs,
        )
        if last_points_only:
            hfcs = [hfcs]

        assert len(hfcs) == 1
        for hfc in hfcs:
            assert not np.isnan(hfc.all_values()).any().any()

        # input too short: without covariates, make `series` shorter. Otherwise, use the shorter covariates
        series_ = series[:-1] if not use_covs else series
        with pytest.raises(ValueError) as exc:
            _ = model.historical_forecasts(
                series=series_,
                **covs_kwargs_short,
                **hfc_kwargs,
            )
        assert str(exc.value).startswith(
            "Could not build the minimum required calibration input with the provided `series` and `*_covariates`"
        )

    @pytest.mark.parametrize("quantiles", [[0.1, 0.5, 0.9], [0.1, 0.3, 0.5, 0.7, 0.9]])
    def test_backtest_and_residuals(self, quantiles):
        """Residuals and backtest are already tested for quantile, and interval metrics based on stochastic or quantile
        forecasts. So, a simple check that they give expected results should be enough.
        """
        n_q = len(quantiles)
        half_idx = n_q // 2
        q_interval = [
            (q_lo, q_hi)
            for q_lo, q_hi in zip(quantiles[:half_idx], quantiles[half_idx + 1 :][::-1])
        ]
        lpo = False

        # series long enough for 2 hfcs
        series = self.helper_prepare_series(True, True).append_values([0.1])
        # conformal model
        model = ConformalNaiveModel(model=train_model(series), quantiles=quantiles)

        hfc = model.historical_forecasts(
            series=series, forecast_horizon=5, last_points_only=lpo, **pred_lklp
        )
        bt = model.backtest(
            series=series,
            historical_forecasts=hfc,
            last_points_only=lpo,
            metric=mic,
            metric_kwargs={"q_interval": model.q_interval},
        )
        # default backtest is equal to backtest with metric kwargs
        np.testing.assert_array_almost_equal(
            bt,
            model.backtest(
                series=series,
                historical_forecasts=hfc,
                last_points_only=lpo,
                metric=mic,
                metric_kwargs={"q_interval": q_interval},
            ),
        )
        np.testing.assert_array_almost_equal(
            mic(
                [series] * len(hfc),
                hfc,
                q_interval=q_interval,
                series_reduction=np.mean,
            ),
            bt,
        )

        residuals = model.residuals(
            series=series,
            historical_forecasts=hfc,
            last_points_only=lpo,
            metric=ic,
            metric_kwargs={"q_interval": q_interval},
        )
        # default residuals is equal to residuals with metric kwargs
        assert residuals == model.residuals(
            series=series,
            historical_forecasts=hfc,
            last_points_only=lpo,
            metric=ic,
            metric_kwargs={"q_interval": q_interval},
        )
        expected_vals = ic([series] * len(hfc), hfc, q_interval=q_interval)
        expected_residuals = []
        for vals, hfc_ in zip(expected_vals, hfc):
            expected_residuals.append(
                TimeSeries.from_times_and_values(
                    times=hfc_.time_index,
                    values=vals,
                    columns=likelihood_component_names(
                        series.components, quantile_interval_names(q_interval)
                    ),
                )
            )
        assert residuals == expected_residuals

    def test_predict_probabilistic_equals_quantile(self):
        """Tests that sampled quantiles predictions have approx. the same quantiles as direct quantile predictions."""
        quantiles = [0.1, 0.3, 0.5, 0.7, 0.9]

        # multiple multivariate series
        series = self.helper_prepare_series(False, False)

        # conformal model
        model = ConformalNaiveModel(model=train_model(series), quantiles=quantiles)
        # direct quantile predictions
        pred_quantiles = model.predict(n=3, series=series, **pred_lklp)
        # sampled predictions
        pred_samples = model.predict(n=3, series=series, num_samples=500)
        for pred_q, pred_s in zip(pred_quantiles, pred_samples):
            assert pred_q.n_samples == 1
            assert pred_q.n_components == series[0].n_components * len(quantiles)
            assert pred_s.n_samples == 500
            assert pred_s.n_components == series[0].n_components

            vals_q = pred_q.all_values()
            vals_s = pred_s.all_values()
            vals_s_q = np.quantile(vals_s, quantiles, axis=2).transpose((1, 2, 0))
            vals_s_q = vals_s_q.reshape(vals_q.shape)
            self.helper_compare_preds(
                vals_s_q,
                vals_q,
                model_type="regression_prob",
            )

    @pytest.mark.parametrize(
        "config",
        [
            # (cal_length, cal_stride, (start_expected, start_format_expected))
            (None, 1, (None, "value")),
            (None, 2, (-4, "position")),
            (None, 3, (-6, "position")),
            (None, 4, (-4, "position")),
            (1, 1, (-3, "position")),
            (1, 2, (-4, "position")),
            (1, 3, (-3, "position")),
            (1, 4, (-4, "position")),
        ],
    )
    def test_calibration_hfc_start_predict(self, config):
        """Test calibration historical forecast start point when calling `predict()` ("end" position)."""
        cal_length, cal_stride, start_expected = config
        series = linear_timeseries(length=4)
        horizon = 2
        output_chunk_shift = 1
        assert (
            _get_calibration_hfc_start(
                series=[series],
                horizon=horizon,
                output_chunk_shift=output_chunk_shift,
                cal_length=cal_length,
                cal_stride=cal_stride,
                start="end",
                start_format="position",
            )
            == start_expected
        )

    @pytest.mark.parametrize(
        "config",
        [
            # (cal_length, cal_stride, start, start_expected)
            (None, 1, None, None),
            (None, 1, 1, None),
            (1, 1, -1, -4),
            (1, 1, 0, 0),
            (1, 2, 0, 0),
            (1, 3, 0, 0),
            (1, 1, 1, 0),
            (1, 2, 1, 1),
            (1, 3, 1, 1),
            (1, 1, -1, -4),
            (1, 2, -1, -5),
            (1, 3, -1, -4),
        ],
    )
    def test_calibration_hfc_start_position_hist_fc(self, config):
        """Test calibration historical forecast start point when calling `historical_forecasts()`
        with start format "position"."""
        cal_length, cal_stride, start, start_expected = config
        series = linear_timeseries(length=4)
        horizon = 2
        output_chunk_shift = 1
        assert _get_calibration_hfc_start(
            series=[series],
            horizon=horizon,
            output_chunk_shift=output_chunk_shift,
            cal_length=cal_length,
            cal_stride=cal_stride,
            start=start,
            start_format="position",
        ) == (start_expected, "position")

    @pytest.mark.parametrize(
        "config",
        [
            # (cal_length, cal_stride, start, start_expected)
            (None, 1, None, None),
            (None, 1, "2020-01-11", None),
            (1, 1, "2020-01-09", "2020-01-06"),  # start before series start
            (1, 1, "2020-01-10", "2020-01-07"),
            (1, 2, "2020-01-10", "2020-01-06"),
            (1, 3, "2020-01-10", "2020-01-07"),
            (2, 1, "2020-01-09", "2020-01-05"),
            (2, 1, "2020-01-10", "2020-01-06"),
            (2, 2, "2020-01-10", "2020-01-04"),
            (2, 3, "2020-01-10", "2020-01-04"),
        ],
    )
    def test_calibration_hfc_start_value_hist_fc(self, config):
        """Test calibration historical forecast start point when calling `historical_forecasts()`
        with start format "value"."""
        cal_length, cal_stride, start, start_expected = config
        if start is not None:
            start = pd.Timestamp(start)
        if start_expected is not None:
            start_expected = pd.Timestamp(start_expected)
        series = linear_timeseries(length=4, start=pd.Timestamp("2020-01-10"), freq="d")
        horizon = 2
        output_chunk_shift = 1
        assert _get_calibration_hfc_start(
            series=[series],
            horizon=horizon,
            output_chunk_shift=output_chunk_shift,
            cal_length=cal_length,
            cal_stride=cal_stride,
            start=start,
            start_format="value",
        ) == (start_expected, "value")

    def test_encoders(self):
        """Tests support of covariates encoders."""
        n = OUT_LEN + 1
        min_length = IN_LEN + n

        # create non-overlapping train and val series
        series = tg.linear_timeseries(length=min_length)
        val_series = tg.linear_timeseries(
            start=series.end_time() + series.freq, length=min_length
        )

        model = train_model(
            series,
            model_params={
                "lags_future_covariates": (IN_LEN, OUT_LEN),
                "add_encoders": {"datetime_attribute": {"future": ["hour"]}},
            },
        )

        cp_model = ConformalNaiveModel(model, quantiles=q)
        assert (
            cp_model.model.encoders is not None
            and cp_model.model.encoders.encoding_available
        )
        assert model.uses_future_covariates

        # predict: encoders using stored train series must work
        _ = cp_model.predict(n=n)
        # predict: encoding of new series without train overlap must work
        _ = cp_model.predict(n=n, series=val_series)

        # check the same for hfc
        _ = cp_model.historical_forecasts(
            forecast_horizon=n, series=series, overlap_end=True
        )
        _ = cp_model.historical_forecasts(
            forecast_horizon=n, series=val_series, overlap_end=True
        )



import logging
import os
import time
import warnings
from typing import NoReturn


def get_logger(name):
    """
    Internally calls the logging.getLogger function with the `name` argument to create or
    retrieve a logger object. It is recommended to pass __name__ as argument when calling
    get_logger. The returned logger object logs to the standard error stream and formats
    the messages appropriately.

    Parameters
    ----------
    name
        The name that gets passed to the logger.getLogger function.

    Returns
    -------
    A logger instance with the given name.
    """

    logger = logging.getLogger(name)

    return logger


def raise_deprecation_warning(
    message: str = "",
    logger: logging.Logger = get_logger("main_logger"),
):
    """
    Raises a DeprecationWarning.

    Parameters
    ----------
    message
        The message of the ValueError.
    logger
        The logger instance to log the error message if 'condition' is False.

    Raises
    ------
    Warning
        DeprecationWarning
    """

    logger.warning("DeprecationWarning: " + message)


def raise_if_not(
    condition: bool,
    message: str = "",
    logger: logging.Logger = get_logger("main_logger"),
):
    """
    Checks provided boolean condition and raises a ValueError if it evaluates to False.
    It logs the error to the provided logger before raising it.

    Parameters
    ----------
    condition
        The boolean condition to be checked.
    message
        The message of the ValueError.
    logger
        The logger instance to log the error message if 'condition' is False.

    Raises
    ------
    ValueError
        if `condition` is not satisfied
    """

    if not condition:
        logger.error("ValueError: " + message)
        raise ValueError(message)


def raise_if(
    condition: bool,
    message: str = "",
    logger: logging.Logger = get_logger("main_logger"),
):
    """
    Checks provided boolean condition and raises a ValueError if it evaluates to True.
    It logs the error to the provided logger before raising it.

    Parameters
    ----------
    condition
        The boolean condition to be checked.
    message
        The message of the ValueError.
    logger
        The logger instance to log the error message if 'condition' is True.

    Raises
    ------
    ValueError
        if `condition` is satisfied
    """
    raise_if_not(not condition, message, logger)


def raise_log(
    exception: Exception, logger: logging.Logger = get_logger("main_logger")
) -> NoReturn:
    """
    Can be used to replace "raise" when throwing an exception to ensure the logging
    of the exception. After logging it, the exception is raised.

    Parameters
    ----------
    exception
        The exception instance to be raised.
    logger
        The logger instance to log the exception type and message.

    Raises
    ------
    Exception
        The provided exception
    """

    exception_type = str(type(exception)).split("'")[1]
    message = str(exception)
    logger.error(exception_type + ": " + message)

    raise exception


def time_log(logger: logging.Logger = get_logger("main_logger")):
    """
    A decorator function that logs the runtime of the function it is decorating
    to the logger object that is taken as an argument.
    Inspired by:
    `https://medium.com/pythonhive/python-decorator-to-measure-the-execution-time-of-methods-fa04cb6bb36d`_.

    Parameters
    ----------
    logger
        The logger instance to log the runtime of the function.
    """

    def time_log_helper(method):
        def timed(*args, **kwargs):
            start_time = time.time()
            result = method(*args, **kwargs)
            end_time = time.time()
            duration = int((end_time - start_time) * 1000)
            logger.info(method.__name__ + f" function ran for {duration} milliseconds")
            return result

        return timed

    return time_log_helper


class SuppressStdoutStderr:
    """
    A context manager for "deep suppression" of stdout and stderr in
    Python, i.e. it suppresses all print, even if the print originates in a
    compiled C/Fortran sub-function.
       This does not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).

    source:
    https://stackoverflow.com/questions/11130156/suppress-stdout-stderr-print-from-python-functions
    """

    def __init__(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)
        # Close all file descriptors
        for fd in self.null_fds + self.save_fds:
            os.close(fd)


def execute_and_suppress_output(function, logger, suppression_threshold_level, *args):
    """
    This function conditionally executes the given function with the given arguments
    based on whether the current level of 'logger' is below, above or equal to
    'suppression_threshold_level'. In the latter two cases, all printouts to stdout
    and stderr will be suppressed.

    :param function: Function whose printouts will conditionally be suppressed.
    :param logger: Logger whose level will be checked.
    :param suppression_threshold_level: Minimum logger level triggering suppression.
    :param *args: Arguments passed to 'function'.
    :return: Outputs of 'function'.
    """
    if logger.level >= suppression_threshold_level:
        with SuppressStdoutStderr():
            return_value = function(*args)
    else:
        return_value = function(*args)
    return return_value


def suppress_lightning_warnings(suppress_all: bool = False):
    warnings.filterwarnings(
        "ignore", ".*You defined a `validation_step` but have no `val_dataloader`.*"
    )
    if suppress_all:
        warnings.filterwarnings(
            "ignore", ".*does not have many workers which may be a bottleneck.*"
        )
        warnings.filterwarnings(
            "ignore",
            ".*Trying to infer the `batch_size` from an ambiguous collection.*",
        )



import numpy as np
import pandas as pd

from darts import TimeSeries
from darts.utils.missing_values import fill_missing_values, missing_values_ratio


class TestMissingValues:
    time = pd.date_range("20130101", "20130130")
    lin = [float(i) for i in range(len(time))]
    cub = [float(i - 4) ** 2 for i in range(len(time))]
    series1: TimeSeries = TimeSeries.from_times_and_values(
        time, np.array([2.0] * len(time))
    )
    series2: TimeSeries = TimeSeries.from_times_and_values(time, np.array(lin))
    series3: TimeSeries = TimeSeries.from_times_and_values(
        time, np.array([10] * 10 + lin[-20:])
    )
    series4: TimeSeries = TimeSeries.from_times_and_values(
        time, np.array(lin[:20] + [19] * 10)
    )
    series5: TimeSeries = TimeSeries.from_times_and_values(time, np.array(cub))
    series6: TimeSeries = TimeSeries.from_times_and_values(
        time, [0] * 2 + cub[2:-2] + [-1] * 2
    )

    def test_fill_constant(self):
        seriesA: TimeSeries = TimeSeries.from_times_and_values(
            self.time,
            np.array(
                [np.nan] * 5 + [2.0] * 5 + [np.nan] * 5 + [2.0] * 10 + [np.nan] * 5
            ),
        )

        # Check that no changes are made if there are no missing values
        assert self.series1 == fill_missing_values(self.series1, "auto")

        # Check that a constant function is filled to a constant function
        assert self.series1 == fill_missing_values(seriesA, "auto")

    def test_linear(self):
        seriesB: TimeSeries = TimeSeries.from_times_and_values(
            self.time, np.array(self.lin[:10] + [np.nan] * 10 + self.lin[-10:])
        )

        # Check for linear interpolation part
        assert self.series2 == fill_missing_values(seriesB, "auto")

    def test_bfill(self):
        seriesC: TimeSeries = TimeSeries.from_times_and_values(
            self.time, np.array([np.nan] * 10 + self.lin[-20:])
        )

        # Check that auto-backfill works properly
        assert self.series3 == fill_missing_values(seriesC, "auto")

    def test_ffil(self):
        seriesD: TimeSeries = TimeSeries.from_times_and_values(
            self.time, np.array(self.lin[:20] + [np.nan] * 10)
        )

        assert self.series4 == fill_missing_values(seriesD, "auto")

    def test_fill_quad(self):
        seriesE: TimeSeries = TimeSeries.from_times_and_values(
            self.time, np.array(self.cub[:10] + [np.nan] * 10 + self.cub[-10:])
        )
        assert self.series5 == round(
            fill_missing_values(seriesE, "auto", method="quadratic"), 7
        )

    def test_multivariate_fill(self):
        seriesA: TimeSeries = TimeSeries.from_times_and_values(
            self.time,
            np.array(
                [np.nan] * 5 + [2.0] * 5 + [np.nan] * 5 + [2.0] * 10 + [np.nan] * 5
            ),
        )
        seriesB: TimeSeries = TimeSeries.from_times_and_values(
            self.time, np.array(self.lin[:10] + [np.nan] * 10 + self.lin[-10:])
        )
        assert self.series1.stack(self.series2) == fill_missing_values(
            seriesA.stack(seriesB), "auto"
        )

    def test_missing_values_ratio(self):
        seriesF = TimeSeries.from_times_and_values(
            self.time, list(range(27)) + [np.nan] * 3
        )

        # univariate case
        assert missing_values_ratio(seriesF) == 0.1

        # multivariate case
        assert missing_values_ratio(seriesF.stack(seriesF)) == 0.1



"""
Explainability Result
--------------------

Contains the explainability results obtained from :func:`_ForecastingModelExplainer.explain()
<darts.explainability.explainability._ForecastingModelExplainer.explain>`.

- :class:`ShapExplainabilityResult <ShapExplainabilityResult>` for :class:`ShapExplainer
  <darts.explainability.shap_explainer.ShapExplainer>`
- :class:`TFTExplainabilityResult <TFTExplainabilityResult>` for :class:`TFTExplainer
  <darts.explainability.tft_explainer.TFTExplainer>`
- :class:`ComponentBasedExplainabilityResult <ComponentBasedExplainabilityResult>` for component based explainability
  results
- :class:`HorizonBasedExplainabilityResult <HorizonBasedExplainabilityResult>` for horizon based explainability results
"""

from abc import ABC, abstractmethod
from typing import Any, Optional, Union

import pandas as pd
import shap

from darts import TimeSeries
from darts.logging import get_logger, raise_if, raise_if_not, raise_log

logger = get_logger(__name__)


class _ExplainabilityResult(ABC):
    """
    Abstract class for explainability results of a :class:`_ForecastingModelExplainer`.
    The subclasses should implement convenient access to explanations.
    """

    @abstractmethod
    def get_explanation(self, *args, **kwargs):
        """Returns one or multiple explanations based on some input parameters."""
        pass


class ComponentBasedExplainabilityResult(_ExplainabilityResult):
    """Explainability result for general component objects.
    The explained components can describe anything.

    Example
    -------
    >>> explainer = SomeComponentBasedExplainer(model)
    >>> explain_results = explainer.explain()
    >>> output = explain_results.get_explanation(component="some_component")
    """

    def __init__(
        self,
        explained_components: Union[dict[str, Any], list[dict[str, Any]]],
    ):
        if isinstance(explained_components, list):
            comps_available = explained_components[0].keys()
            if not all(comp.keys() == comps_available for comp in explained_components):
                raise_log(
                    ValueError(
                        "When giving a list of explained component dicts, the dict keys must match."
                    ),
                    logger=logger,
                )
        else:
            comps_available = explained_components.keys()
        self.explained_components = explained_components
        self.available_components = comps_available

    def get_explanation(self, component) -> Union[Any, list[Any]]:
        """
        Returns one or several explanations for a given component.

        Parameters
        ----------
        component
            The component for which to return the explanation.
        """
        return self._query_explainability_result(self.explained_components, component)

    def _query_explainability_result(
        self,
        attr: Union[dict[str, Any], list[dict[str, Any]]],
        component: str,
    ) -> Any:
        """
        Helper that extracts and returns the explainability result attribute for a given component.

        Parameters
        ----------
        attr
            An explainability result attribute from which to extract the component.
        component
            The component for which to return the content of the attribute.
        """
        component = self._validate_input_for_querying_explainability_result(component)
        if isinstance(attr, list):
            return [attr_[component] for attr_ in attr]
        else:
            return attr[component]

    def _validate_input_for_querying_explainability_result(self, component) -> str:
        """
        Helper that validates the input parameters of a method that queries the `ComponentBasedExplainabilityResult`.

        Parameters
        ----------
        component
            The component for which to return the explanation. Does not
            need to be specified for univariate series.
        """
        # validate component argument
        raise_if(
            component is None and len(self.explained_components) > 1,
            f"The component parameter is required when the `{self.__class__.__name__}` has more than one component.",
            logger,
        )

        if component is None:
            component = self.available_components[0]

        raise_if_not(
            component in self.available_components,
            f"Component {component} is not available. Available components are: {self.available_components}",
            logger,
        )
        return component


class HorizonBasedExplainabilityResult(_ExplainabilityResult):
    """
    Stores the explainability results of a :class:`_ForecastingModelExplainer
    <darts.explainability.explainability._ForecastingModelExplainer>` with convenient access to the horizon
    based results.

    The result is a multivariate `TimeSeries` instance containing the 'explanation' for the (horizon, target_component)
    forecast at any timestamp forecastable corresponding to the foreground `TimeSeries` input.

    The component name convention of this multivariate `TimeSeries` is:
    ``"{name}_{type_of_cov}_lag_{idx}"``, where:

    - ``{name}`` is the component name from the original foreground series (target, past, or future).
    - ``{type_of_cov}`` is the covariates type. It can take 3 different values:
      ``"target"``, ``"past_cov"`` or ``"future_cov"``.
    - ``{idx}`` is the lag index.

    Example
    -------

    Say we have a model with 2 target components named ``"T_0"`` and ``"T_1"``,
    3 past covariates with default component names ``"0"``, ``"1"``, and ``"2"``,
    and one future covariate with default component name ``"0"``.
    Also, ``horizons = [1, 2]``.
    The model is a regression model, with ``lags = 3``, ``lags_past_covariates=[-1, -3]``,
    ``lags_future_covariates = [0]``.

    We provide `foreground_series`, `foreground_past_covariates`, `foreground_future_covariates` each of length 5.

    >>> explainer = SomeHorizonBasedExplainer(model)
    >>> explain_results = explainer.explain(
    >>>     foreground_series=foreground_series,
    >>>     foreground_past_covariates=foreground_past_covariates,
    >>>     foreground_future_covariates=foreground_future_covariates,
    >>>     horizons=[1, 2],
    >>>     target_names=["T_0", "T_1"]
    >>> )
    >>> output = explain_results.get_explanation(horizon=1, target="T_1")

    Then the method returns a multivariate TimeSeries containing the *explanations* of
    the corresponding `_ForecastingModelExplainer`, with the following component names:

         - T_0_target_lag-1
         - T_0_target_lag-2
         - T_0_target_lag-3
         - T_1_target_lag-1
         - T_1_target_lag-2
         - T_1_target_lag-3
         - 0_past_cov_lag-1
         - 0_past_cov_lag-3
         - 1_past_cov_lag-1
         - 1_past_cov_lag-3
         - 2_past_cov_lag-1
         - 2_past_cov_lag-3
         - 0_fut_cov_lag_0

    This series has length 3, as the model can explain 5-3+1 forecasts
    (timestamp indexes 4, 5, and 6)
    """

    def __init__(
        self,
        explained_forecasts: Union[
            dict[int, dict[str, TimeSeries]],
            list[dict[int, dict[str, TimeSeries]]],
        ],
    ):
        self.explained_forecasts = explained_forecasts
        if isinstance(self.explained_forecasts, list):
            raise_if_not(
                isinstance(self.explained_forecasts[0], dict),
                "The explained_forecasts list must consist of dicts.",
                logger,
            )
            raise_if_not(
                all(isinstance(key, int) for key in self.explained_forecasts[0].keys()),
                "The explained_forecasts dict list must have all integer keys.",
                logger,
            )
            self.available_horizons = list(self.explained_forecasts[0].keys())
            h_0 = self.available_horizons[0]
            self.available_components = list(self.explained_forecasts[0][h_0].keys())
        elif isinstance(self.explained_forecasts, dict):
            if all(isinstance(key, int) for key in self.explained_forecasts.keys()):
                self.available_horizons = list(self.explained_forecasts.keys())
                h_0 = self.available_horizons[0]
                self.available_components = list(self.explained_forecasts[h_0].keys())
            else:
                raise_log(
                    ValueError(
                        "The explained_forecasts dictionary must have all integer keys."
                    ),
                    logger,
                )
        else:
            raise_log(
                ValueError(
                    "The explained_forecasts must be a dictionary or a list of dictionaries."
                ),
                logger,
            )

    def get_explanation(
        self, horizon: int, component: Optional[str] = None
    ) -> Union[TimeSeries, list[TimeSeries]]:
        """
        Returns one or several `TimeSeries` representing the explanations
        for a given horizon and component.

        Parameters
        ----------
        horizon
            The horizon for which to return the explanation.
        component
            The component for which to return the explanation. Does not
            need to be specified for univariate series.
        """
        return self._query_explainability_result(
            self.explained_forecasts, horizon, component
        )

    def _query_explainability_result(
        self,
        attr: Union[dict[int, dict[str, Any]], list[dict[int, dict[str, Any]]]],
        horizon: int,
        component: Optional[str] = None,
    ) -> Any:
        """
        Helper that extracts and returns the explainability result attribute for a specified horizon and component from
        the input attribute.

        Parameters
        ----------
        attr
            An explainability result attribute from which to extract the content for a certain horizon and component.
        horizon
            The horizon for which to return the content of the attribute.
        component
            The component for which to return the content of the attribute. Does not
            need to be specified for univariate series.
        """
        component = self._validate_input_for_querying_explainability_result(
            horizon, component
        )

        if isinstance(attr, list):
            return [attr[i][horizon][component] for i in range(len(attr))]
        elif all(isinstance(key, int) for key in attr.keys()):
            return attr[horizon][component]
        else:
            raise_log(
                ValueError(
                    f"Something went wrong. {self.__class__.__name__} got instantiated with an unexpected type."
                ),
                logger,
            )

    def _validate_input_for_querying_explainability_result(
        self, horizon: int, component: Optional[str] = None
    ) -> str:
        """
        Helper that validates the input parameters of a method that queries the `HorizonBasedExplainabilityResult`.

        Parameters
        ----------
        horizon
            The horizon for which to return the explanation.
        component
            The component for which to return the explanation. Does not
            need to be specified for univariate series.
        """
        # validate component argument
        raise_if(
            component is None and len(self.available_components) > 1,
            "The component parameter is required when the model has more than one component.",
            logger,
        )

        if component is None:
            component = self.available_components[0]

        raise_if_not(
            component in self.available_components,
            f"Component {component} is not available. Available components are: {self.available_components}",
            logger,
        )

        raise_if_not(
            horizon in self.available_horizons,
            f"Horizon {horizon} is not available. Available horizons are: {self.available_horizons}",
            logger,
        )
        return component


class ShapExplainabilityResult(HorizonBasedExplainabilityResult):
    """
    Stores the explainability results of a :class:`ShapExplainer <darts.explainability.shap_explainer.ShapExplainer>`
    with convenient access to the results. It extends the :class:`HorizonBasedExplainabilityResult
    <HorizonBasedExplainabilityResult>` and carries additional information specific to the Shap explainers.
    In particular, in addition to the `explained_forecasts` (which in the case of the `ShapExplainer` are the
    shap values), it also provides access to the corresponding `feature_values` and the underlying `shap.Explanation`
    object.

    - :func:`get_explanation() <ShapExplainabilityResult.get_explanation>`: explained forecast for a given horizon
      (and target component)
    - :func:`get_feature_values() <ShapExplainabilityResult.get_feature_values>`: feature values for a given horizon
      (and target component).
    - :func:`get_shap_explanation_object() <ShapExplainabilityResult.get_shap_explanation_object>`: `shap.Explanation`
      object for a given horizon (and target component).

    Examples
    --------
    >>> explainer = ShapExplainer(model)  # requires `background` if model was trained on multiple series
    >>> explain_results = explainer.explain()
    >>> exlained_fc = explain_results.get_explanation(horizon=1)
    >>> feature_values = explain_results.get_feature_values(horizon=1)
    >>> shap_objects = explain_results.get_shap_explanation_objects(horizon=1)
    """

    def __init__(
        self,
        explained_forecasts: Union[
            dict[int, dict[str, TimeSeries]],
            list[dict[int, dict[str, TimeSeries]]],
        ],
        feature_values: Union[
            dict[int, dict[str, TimeSeries]],
            list[dict[int, dict[str, TimeSeries]]],
        ],
        shap_explanation_object: Union[
            dict[int, dict[str, shap.Explanation]],
            list[dict[int, dict[str, shap.Explanation]]],
        ],
    ):
        super().__init__(explained_forecasts)
        self.feature_values = feature_values
        self.shap_explanation_object = shap_explanation_object

    def get_feature_values(
        self, horizon: int, component: Optional[str] = None
    ) -> Union[TimeSeries, list[TimeSeries]]:
        """
        Returns one or several `TimeSeries` representing the feature values
        for a given horizon and component.

        Parameters
        ----------
        horizon
            The horizon for which to return the feature values.
        component
            The component for which to return the feature values. Does not
            need to be specified for univariate series.
        """
        return self._query_explainability_result(
            self.feature_values, horizon, component
        )

    def get_shap_explanation_object(
        self, horizon: int, component: Optional[str] = None
    ) -> Union[shap.Explanation, list[shap.Explanation]]:
        """
        Returns the underlying `shap.Explanation` object for a given horizon and component.

        Parameters
        ----------
        horizon
            The horizon for which to return the `shap.Explanation` object.
        component
            The component for which to return the `shap.Explanation` object. Does not
            need to be specified for univariate series.
        """
        return self._query_explainability_result(
            self.shap_explanation_object, horizon, component
        )


class TFTExplainabilityResult(ComponentBasedExplainabilityResult):
    """
    Stores the explainability results of a :class:`TFTExplainer <darts.explainability.tft_explainer.TFTExplainer>`
    with convenient access to the results. It extends the :class:`ComponentBasedExplainabilityResult` and carries
    information specific to the TFT explainer.

    - :func:`get_attention() <TFTExplainabilityResult.get_attention>`: self attention over the encoder and decoder
    - :func:`get_encoder_importance() <TFTExplainabilityResult.get_encoder_importance>`: encoder feature importances
      including past target, past covariates, and historic part of future covariates.
    - :func:`get_decoder_importance() <TFTExplainabilityResult.get_decoder_importance>`: decoder feature importances
      including future part of future covariates.
    - :func:`get_static_covariates_importance() <TFTExplainabilityResult.get_static_covariates_importance>`: static
      covariates importances.
    - :func:`get_feature_importances() <TFTExplainabilityResult.get_feature_importances>`: get all feature importances
      at once.

    Examples
    --------
    >>> explainer = TFTExplainer(model)  # requires `background` if model was trained on multiple series
    >>> explain_results = explainer.explain()
    >>> attention = explain_results.get_attention()
    >>> importances = explain_results.get_feature_importances()
    >>> encoder_importance = explain_results.get_encoder_importance()
    >>> decoder_importance = explain_results.get_decoder_importance()
    >>> static_covariates_importance = explain_results.get_static_covariates_importance()
    """

    def __init__(
        self,
        explanations: Union[
            dict[str, Any],
            list[dict[str, Any]],
        ],
    ):
        super().__init__(explanations)
        self.feature_importances = [
            "encoder_importance",
            "decoder_importance",
            "static_covariates_importance",
        ]

    def get_attention(self) -> Union[TimeSeries, list[TimeSeries]]:
        """
        Returns the time-dependent attention on the encoder and decoder for each `horizon` in (1,
        `output_chunk_length`). The time index ranges from the prediction series' start time - input_chunk_length and
        ends at the prediction series' end time. If multiple series were used when calling
        :func:`TFTExplainer.explain() <darts.explainability.tft_explainer.TFTExplainer.explain>`, returns a list of
        TimeSeries.
        """
        attention = self.get_explanation("attention")
        return attention

    def get_feature_importances(
        self,
    ) -> dict[str, Union[pd.DataFrame, list[pd.DataFrame]]]:
        """
        Returns the feature importances for the encoder, decoder and static covariates as pd.DataFrames.
        If multiple series were used in :func:`TFTExplainer.explain()
        <darts.explainability.tft_explainer.TFTExplainer.explain>`, returns a list of pd.DataFrames per importance.
        """
        return {comp: self.get_explanation(comp) for comp in self.feature_importances}

    def get_encoder_importance(self) -> Union[pd.DataFrame, list[pd.DataFrame]]:
        """
        Returns the time-dependent encoder importances as a pd.DataFrames.
        If multiple series were used in :func:`TFTExplainer.explain()
        <darts.explainability.tft_explainer.TFTExplainer.explain>`, returns a list of pd.DataFrames.
        """
        return self.get_explanation("encoder_importance")

    def get_decoder_importance(self) -> Union[pd.DataFrame, list[pd.DataFrame]]:
        """
        Returns the time-dependent decoder importances as a pd.DataFrames.
        If multiple series were used in :func:`TFTExplainer.explain()
        <darts.explainability.tft_explainer.TFTExplainer.explain>`, returns a list of pd.DataFrames.
        """
        return self.get_explanation("decoder_importance")

    def get_static_covariates_importance(
        self,
    ) -> Union[pd.DataFrame, list[pd.DataFrame]]:
        """
        Returns the numeric and categorical static covariates importances as a pd.DataFrames.
        If multiple series were used in :func:`TFTExplainer.explain()
        <darts.explainability.tft_explainer.TFTExplainer.explain>`, returns a list of pd.DataFrames.
        """
        return self.get_explanation("static_covariates_importance")



from typing import Optional

from sklearn import __version__ as sklearn_version
from sklearn.base import is_classifier
from sklearn.multioutput import MultiOutputRegressor as sk_MultiOutputRegressor
from sklearn.multioutput import _fit_estimator
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import has_fit_parameter

from darts.logging import get_logger, raise_log

if sklearn_version >= "1.4":
    # sklearn renamed `_check_fit_params` to `_check_method_params` in v1.4
    from sklearn.utils.validation import _check_method_params
else:
    from sklearn.utils.validation import _check_fit_params as _check_method_params

if sklearn_version >= "1.3":
    # delayed was moved from sklearn.utils.fixes to sklearn.utils.parallel in v1.3
    from sklearn.utils.parallel import Parallel, delayed
else:
    from joblib import Parallel
    from sklearn.utils.fixes import delayed

logger = get_logger(__name__)


class MultiOutputRegressor(sk_MultiOutputRegressor):
    """
    :class:`sklearn.utils.multioutput.MultiOutputRegressor` with a modified ``fit()`` method that also slices
    validation data correctly. The validation data has to be passed as parameter ``eval_set`` in ``**fit_params``.
    """

    def __init__(
        self,
        *args,
        eval_set_name: Optional[str] = None,
        eval_weight_name: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.eval_set_name_ = eval_set_name
        self.eval_weight_name_ = eval_weight_name
        self.estimators_ = None
        self.n_features_in_ = None
        self.feature_names_in_ = None

    def fit(self, X, y, sample_weight=None, **fit_params):
        """Fit the model to data, separately for each output variable.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input data.

        y : {array-like, sparse matrix} of shape (n_samples, n_outputs)
            Multi-output targets. An indicator matrix turns on multilabel
            estimation.

        sample_weight : array-like of shape (n_samples, n_outputs), default=None
            Sample weights. If `None`, then samples are equally weighted.
            Only supported if the underlying regressor supports sample
            weights.

        **fit_params : dict of string -> object
            Parameters passed to the ``estimator.fit`` method of each step.

            .. versionadded:: 0.23

        Returns
        -------
        self : object
            Returns a fitted instance.
        """

        if not hasattr(self.estimator, "fit"):
            raise_log(
                ValueError("The base estimator should implement a fit method"),
                logger=logger,
            )

        y = self._validate_data(X="no_validation", y=y, multi_output=True)

        if is_classifier(self):
            check_classification_targets(y)

        if y.ndim == 1:
            raise_log(
                ValueError(
                    "`y` must have at least two dimensions for multi-output regression but has only one."
                ),
                logger=logger,
            )
        if sample_weight is not None and (
            sample_weight.ndim == 1 or sample_weight.shape[1] != y.shape[1]
        ):
            raise_log(
                ValueError("`sample_weight` must have the same dimensions as `y`."),
                logger=logger,
            )

        if sample_weight is not None and not self.supports_sample_weight:
            raise_log(
                ValueError("Underlying estimator does not support sample weights."),
                logger=logger,
            )

        fit_params_validated = _check_method_params(X, fit_params)
        eval_set = fit_params_validated.pop(self.eval_set_name_, None)
        eval_weight = fit_params_validated.pop(self.eval_weight_name_, None)

        self.estimators_ = Parallel(n_jobs=self.n_jobs)(
            delayed(_fit_estimator)(
                self.estimator,
                X,
                y[:, i],
                sample_weight=sample_weight[:, i]
                if sample_weight is not None
                else None,
                **(
                    {self.eval_set_name_: [eval_set[i]]} if eval_set is not None else {}
                ),
                **(
                    {self.eval_weight_name_: [eval_weight[i]]}
                    if eval_weight is not None
                    else {}
                ),
                **fit_params_validated,
            )
            for i in range(y.shape[1])
        )

        if hasattr(self.estimators_[0], "n_features_in_"):
            self.n_features_in_ = self.estimators_[0].n_features_in_
        if hasattr(self.estimators_[0], "feature_names_in_"):
            self.feature_names_in_ = self.estimators_[0].feature_names_in_

        return self

    @property
    def supports_sample_weight(self) -> bool:
        """
        Whether model supports sample weight for training.
        """
        return has_fit_parameter(self.estimator, "sample_weight")
